from torch import nn

from .activation import BoundReLU, BoundTanh, BoundSigmoid, BoundIdentity, BoundExp, Exp, BoundLog, Log, \
    BoundReciprocal, Reciprocal, BoundSin, Sin, BoundCos, Cos
from .bivariate import BoundAdd, Add, BoundSub, Sub, VectorAdd, BoundVectorAdd, VectorSub, BoundVectorSub, VectorMul, \
    BoundVectorMul, BoundMul, Mul
from .linear import BoundLinear, BoundElementWiseLinear, ElementWiseLinear
from .parallel import BoundParallel, Parallel
from .polynomial import BoundPow, Pow
from .probability import BoundErf, Erf, StandardNormalPDF, BoundStandardNormalPDF
from .reshape import BoundSelect, Select, BoundFlip, Flip
from .saturation import BoundClamp, Clamp
from .sequential import BoundSequential


class BoundModelFactory:
    """
    This class is a combination of the factory and visitor pattern to construct the bound module graph recursively while
    being flexible to add new types.
    As an example, say you design a new nn.Module subtype and let that be a layer in a nn.Sequential. With the factory,
    you can add the new subtype to the factory, and when constructing BoundSequential, this will be recognized.
    """
    def __init__(self, **kwargs):
        self.class_mapping = [
            (nn.Sequential, BoundSequential),
            (nn.Linear, BoundLinear),
            (ElementWiseLinear, BoundElementWiseLinear),
            (nn.ReLU, BoundReLU),
            (nn.Tanh, BoundTanh),
            (nn.Sigmoid, BoundSigmoid),
            (nn.Identity, BoundIdentity),
            (Exp, BoundExp),
            (Log, BoundLog),
            (Reciprocal, BoundReciprocal),
            (Sin, BoundSin),
            (Cos, BoundCos),
            (Add, BoundAdd),
            (VectorAdd, BoundVectorAdd),
            (Sub, BoundSub),
            (VectorSub, BoundVectorSub),
            (Mul, BoundMul),
            (VectorMul, BoundVectorMul),
            (Parallel, BoundParallel),
            (Clamp, BoundClamp),
            (Select, BoundSelect),
            (Flip, BoundFlip),
            (Pow, BoundPow),
            (Erf, BoundErf),
            (StandardNormalPDF, BoundStandardNormalPDF)
        ]

        self.kwargs = kwargs

    def register(self, original_class, bound_class):
        self.class_mapping.insert(0, (original_class, bound_class))

    def build(self, module):
        for orig_class, bound_class in self.class_mapping:
            if isinstance(module, orig_class):
                return bound_class(module, self, **self.kwargs)

        raise NotImplementedError('Module type not supported - add BoundModule for layer to factory')



from typing import Callable

import torch


# For convenience in specifying type hints and for a semantic name (understanding)
TensorFunction = Callable[[torch.Tensor], torch.Tensor]


def proj_grad_to_range_(param, range):
    param.grad = param.data - (param.data - param.grad).clamp(min=range[0], max=range[1])


def clip_param_to_range_(param, range):
    param.data.clamp_(min=range[0], max=range[1])



import torch
from torch import nn

from .general import BoundModule
from .bounds import IntervalBounds, LinearBounds


class Select(nn.Module):
    def __init__(self, indices):
        super().__init__()

        if isinstance(indices, int):
            indices = [indices]
        elif isinstance(indices, slice):
            if indices.step is None:
                r = range(indices.start, indices.stop)
            else:
                r = range(indices.start, indices.stop, indices.step)

            indices = [i for i in r]

        self.indices = indices

    def forward(self, x):
        return x[..., self.indices]


class BoundSelect(BoundModule):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory)

        self.in_size = None

    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        assert self.in_size

        if linear_bounds.lower is None:
            lower = None
        else:
            lowerA = torch.zeros((*linear_bounds.lower[0].size()[:-1], self.in_size), device=linear_bounds.lower[0].device)

            view_size = *[1 for _ in range(linear_bounds.lower[0].dim() - 1)], -1
            indices = torch.tensor(self.module.indices, device=linear_bounds.lower[0].device).view(*view_size).expand_as(linear_bounds.lower[0])
            lowerA.scatter_add_(-1, indices, linear_bounds.lower[0])

            lower = (lowerA, linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upperA = torch.zeros((*linear_bounds.upper[0].size()[:-1], self.in_size), device=linear_bounds.upper[0].device)

            view_size = *[1 for _ in range(linear_bounds.upper[0].dim() - 1)], -1
            indices = torch.tensor(self.module.indices, device=linear_bounds.upper[0].device).view(*view_size).expand_as(linear_bounds.upper[0])
            upperA.scatter_add_(-1, indices, linear_bounds.upper[0])
            upper = (upperA, linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        return IntervalBounds(bounds.region, bounds.lower[..., self.module.indices], bounds.upper[..., self.module.indices])

    def propagate_size(self, in_size):
        self.in_size = in_size

        return len(self.module.indices)


class Flip(nn.Module):
    def forward(self, x):
        assert x.size(-1) % 2 == 0
        size = x.size(-1) // 2

        return torch.cat([x[..., size:], x[..., :size]], dim=-1)


class BoundFlip(BoundModule):
    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        if linear_bounds.lower is None:
            lower = None
        else:
            assert linear_bounds.lower[0].size(-1) % 2 == 0
            size = linear_bounds.lower[0].size(-1) // 2

            lowerA = torch.cat([linear_bounds.lower[0][..., size:], linear_bounds.lower[0][..., :size]], dim=-1)
            lower = (lowerA, linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            assert linear_bounds.upper[0].size(-1) % 2 == 0
            size = linear_bounds.upper[0].size(-1) // 2

            upperA = torch.cat([linear_bounds.upper[0][..., size:], linear_bounds.upper[0][..., :size]], dim=-1)
            upper = (upperA, linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        return IntervalBounds(bounds.region, self(bounds.lower), self(bounds.upper))

    def propagate_size(self, in_size):
        return in_size



from .activation import BoundActivation, BoundTanh, BoundReLU, BoundSigmoid, bisection, Exp, BoundExp, Log, BoundLog, \
    Reciprocal, BoundReciprocal, Sin, BoundSin, Cos, BoundCos
from .bounds import LinearBounds, IntervalBounds, HyperRectangle
from .factory import BoundModelFactory
from .general import BoundModule
from .linear import BoundLinear, FixedLinear, ElementWiseLinear, BoundElementWiseLinear
from .sequential import BoundSequential
from .parallel import Parallel, BoundParallel, Cat
from .bivariate import Add, BoundAdd, VectorAdd, BoundVectorAdd, Sub, BoundSub, VectorSub, BoundVectorSub, Residual, Mul, BoundMul, Div, VectorMul, BoundVectorMul
from .saturation import Clamp, BoundClamp
from .reshape import Select, BoundSelect, Flip, BoundFlip
from .polynomial import Pow, BoundPow, UnivariateMonomial, MultivariateMonomial
from .probability import Erf, BoundErf, StandardNormalPDF, BoundStandardNormalPDF, StandardNormalCDF, NormalPDF, NormalCDF, \
    TruncatedGaussianTwoSidedExpectation, TruncatedGaussianLowerTailExpectation, TruncatedGaussianUpperTailExpectation



import torch
from torch import nn

from .bounds import IntervalBounds, LinearBounds
from .general import BoundModule


class Parallel(nn.Module):
    def __init__(self, *subnetworks, split_size=None):
        super().__init__()

        self.split_size = split_size
        self.subnetworks = nn.ModuleList(subnetworks)

    def forward(self, x):
        if self.split_size is not None:
            x = x.split(self.split_size, dim=-1)
            y = [network(x) for network, x in zip(self.subnetworks, x)]
        else:
            y = [network(x) for network in self.subnetworks]

        return torch.cat(y, dim=-1)


class BoundParallel(BoundModule):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory)

        self.in_sizes = None
        self.out_sizes = None
        self.subnetworks = nn.ModuleList([factory.build(network) for network in module.subnetworks])

    @property
    def need_relaxation(self):
        return any([network.need_relaxation for network in self.subnetworks])

    def clear_relaxation(self):
        for network in self.subnetworks:
            network.clear_relaxation()

    def set_relaxation(self, linear_bounds, modules, sizes, extras):
        indices = self.indices(sizes)
        for module, index, extra in zip(modules, indices, extras):
            module.set_relaxation(linear_bounds[..., index[0]:index[1], :], *extra)

    def backward_relaxation(self, region):
        assert any([network.need_relaxation for network in self.subnetworks])
        bounds = [self.submodule_backward_relaxation(region, network) for network in self.subnetworks]

        if self.module.split_size is None:
            bounds = [bound for bound in bounds if bound[1] is not None]

            modules = [bound[1] for bound in bounds]
            sizes = [bound[0].lower[0].size(-2) for bound in bounds]

            lowerA = torch.cat([bound[0].lower[0] for bound in bounds], dim=-2)
            lower_bias = torch.cat([bound[0].lower[1] for bound in bounds], dim=-1)

            upperA = torch.cat([bound[0].upper[0] for bound in bounds], dim=-2)
            upper_bias = torch.cat([bound[0].upper[1] for bound in bounds], dim=-1)

            extras = [bound[2:] for bound in bounds]
            return LinearBounds(region, (lowerA, lower_bias), (upperA, upper_bias)), self, modules, sizes, extras
        else:
            bounds = [self.submodule_backward_relaxation(region, network) for network in self.subnetworks]

            indices = self.indices(self.in_sizes)
            bounds = [(bound, index) for bound, index in zip(bounds, indices) if bound[1] is not None]

            modules = [bound[1] for bound, index in bounds]
            sizes = [bound[0].lower[0].size(-2) for bound, index in bounds]
            total_size = sum(self.in_sizes)

            lowerA = torch.cat([self.padding(bound[0].lower[0], index, total_size) for bound, index in bounds], dim=-2)
            lower_bias = torch.cat([bound[0].lower[1] for bound, index in bounds], dim=-1)

            upperA = torch.cat([self.padding(bound[0].upper[0], index, total_size) for bound, index in bounds], dim=-2)
            upper_bias = torch.cat([bound[0].upper[1] for bound, index in bounds], dim=-1)

            extras = [bound[2:] for bound, index in bounds]
            return LinearBounds(region, (lowerA, lower_bias), (upperA, upper_bias)), self, modules, sizes, extras

    def submodule_backward_relaxation(self, region, network):
        if network.need_relaxation:
            return network.backward_relaxation(region)
        else:
            return None, None

    def padding(self, A, index, total_size):
        size = list(A.size())
        size[-1] = index[0]
        pre_padding = torch.zeros(size, device=A.device)

        size[-1] = total_size - index[1]
        post_padding = torch.zeros(size, device=A.device)

        return torch.cat([pre_padding, A, post_padding], dim=-1)

    def crown_backward(self, linear_bounds, optimize):
        assert self.out_sizes is not None
        split_bounds = self.split(linear_bounds, self.out_sizes)
        residual_bounds = [network.crown_backward(split_linear_bound, optimize) for network, split_linear_bound in zip(self.subnetworks, split_bounds)]

        if linear_bounds.lower is None:
            lower = None
        else:
            if self.module.split_size is None:
                lowerA = torch.stack([residual_linear_bound.lower[0] for residual_linear_bound in residual_bounds], dim=-1).sum(dim=-1)
            else:
                lowerA = torch.cat([residual_linear_bound.lower[0] for residual_linear_bound in residual_bounds], dim=-1)

            lower = (lowerA, torch.stack([residual_linear_bound.lower[1] for residual_linear_bound in residual_bounds], dim=-1).sum(dim=-1) - (len(self.subnetworks) - 1) * linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            if self.module.split_size is None:
                upperA = torch.stack([residual_linear_bound.upper[0] for residual_linear_bound in residual_bounds], dim=-1).sum(dim=-1)
            else:
                upperA = torch.cat([residual_linear_bound.upper[0] for residual_linear_bound in residual_bounds], dim=-1)

            upper = (upperA, torch.stack([residual_linear_bound.upper[1] for residual_linear_bound in residual_bounds], dim=-1).sum(dim=-1) - (len(self.subnetworks) - 1) * linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        if self.module.split_size is None:
            split_bounds = [bounds for _ in range(len(self.subnetworks))]
        else:
            split_bounds = self.split(bounds, self.split_sizes(bounds.lower.size(-1)))

        residual_bounds = [network.ibp_forward(bound, save_relaxation=save_relaxation, save_input_bounds=save_input_bounds) for network, bound in zip(self.subnetworks, split_bounds)]

        lower = torch.cat([bound.lower for bound in residual_bounds], dim=-1)
        upper = torch.cat([bound.upper for bound in residual_bounds], dim=-1)
        return IntervalBounds(bounds.region, lower, upper)

    def split(self, bounds, sizes):
        indices = self.indices(sizes)
        return [bounds[..., from_:to_] for from_, to_ in indices]

    def indices(self, sizes):
        indices = torch.tensor(sizes).cumsum(0).tolist()
        return list(zip([0] + indices[:-1], indices))

    def propagate_size(self, in_size):
        split_sizes = self.module.split_size
        if split_sizes is not None:
            split_sizes = self.split_sizes(in_size)

            out_sizes = [network.propagate_size(split_size) for network, split_size in zip(self.subnetworks, split_sizes)]
        else:
            out_sizes = [network.propagate_size(in_size) for network in self.subnetworks]

        self.out_sizes = out_sizes
        self.in_sizes = split_sizes

        return sum(out_sizes)

    def split_sizes(self, in_size):
        split_sizes = self.module.split_size
        if isinstance(split_sizes, int):
            num_full_size = in_size // split_sizes
            split_sizes = [split_sizes for _ in range(num_full_size)]
            if in_size % self.module.split_size != 0:
                split_sizes.append(in_size - num_full_size * self.module.split_size)

        return split_sizes

    def bound_parameters(self):
        for network in self.subnetworks:
            yield from network.bound_parameters()

    def clip_params(self):
        for network in self.subnetworks:
            network.clip_params()

    def project_grads(self):
        for network in self.subnetworks:
            network.project_grads()


class Cat(Parallel):
    def __init__(self, subnetwork):
        super().__init__(
            nn.Identity(),
            subnetwork
        )



from torch import nn

from bound_propagation.general import BoundModule


class BoundSequential(BoundModule):
    def __init__(self, sequential, factory, **kwargs):
        super().__init__(sequential, factory, **kwargs)
        self.bound_sequential = nn.Sequential(*list(map(factory.build, sequential)))

    @property
    def need_relaxation(self):
        return any((module.need_relaxation for module in self.bound_sequential))

    def clear_relaxation(self):
        for module in self.bound_sequential:
            module.clear_relaxation()

    def backward_relaxation(self, region):
        for i, module in enumerate(self.bound_sequential):
            if module.need_relaxation:
                linear_bounds, relaxation_module, *extra = module.backward_relaxation(region)
                return self.subnetwork_crown_backward(self.bound_sequential[:i], linear_bounds, False), relaxation_module, *extra

        assert False, 'At least one module needs relaxation'

    def subnetwork_crown_backward(self, subnetwork, linear_bounds, optimize):
        for module in list(reversed(subnetwork)):
            linear_bounds = module.crown_backward(linear_bounds, optimize)

        return linear_bounds

    def crown_backward(self, linear_bounds, optimize):
        return self.subnetwork_crown_backward(self.bound_sequential, linear_bounds, optimize)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        for module in self.bound_sequential:
            bounds = module.ibp_forward(bounds, save_relaxation=save_relaxation, save_input_bounds=save_input_bounds)

        return bounds

    def propagate_size(self, in_size):
        size = in_size
        for module in self.bound_sequential:
            size = module.propagate_size(size)

        return size

    def bound_parameters(self):
        for module in self.bound_sequential:
            yield from module.bound_parameters()

    def clip_params(self):
        for module in self.bound_sequential:
            module.clip_params()

    def project_grads(self):
        for module in self.bound_sequential:
            module.project_grads()



import logging
from typing import Tuple

import torch
from torch import nn
from torch.nn import Identity

from .activation import Reciprocal
from .general import BoundModule
from .bounds import IntervalBounds, LinearBounds
from .util import clip_param_to_range_, proj_grad_to_range_

logger = logging.getLogger(__name__)


class Add(nn.Module):
    def __init__(self, network1, network2):
        super().__init__()

        self.network1, self.network2 = network1, network2

    def forward(self, x):
        return self.network1(x) + self.network2(x)


class BoundAdd(BoundModule):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.bound_network1 = factory.build(module.network1)
        self.bound_network2 = factory.build(module.network2)

    @property
    def need_relaxation(self):
        return self.bound_network1.need_relaxation or self.bound_network2.need_relaxation

    def clear_relaxation(self):
        self.bound_network1.clear_relaxation()
        self.bound_network2.clear_relaxation()

    def backward_relaxation(self, region):
        if self.bound_network1.need_relaxation:
            return self.bound_network1.backward_relaxation(region)
        else:
            assert self.bound_network2.need_relaxation
            return self.bound_network2.backward_relaxation(region)

    def crown_backward(self, linear_bounds, optimize):
        linear_bounds1 = self.bound_network1.crown_backward(linear_bounds, optimize)
        linear_bounds2 = self.bound_network2.crown_backward(linear_bounds, optimize)

        if linear_bounds.lower is None:
            lower = None
        else:
            lower = (linear_bounds1.lower[0] + linear_bounds2.lower[0], linear_bounds1.lower[1] + linear_bounds2.lower[1] - linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = (linear_bounds1.upper[0] + linear_bounds2.upper[0], linear_bounds1.upper[1] + linear_bounds2.upper[1] - linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        bounds1 = self.bound_network1.ibp_forward(bounds, save_relaxation=save_relaxation, save_input_bounds=save_input_bounds)
        bounds2 = self.bound_network2.ibp_forward(bounds, save_relaxation=save_relaxation, save_input_bounds=save_input_bounds)

        return IntervalBounds(
            bounds.region,
            bounds1.lower + bounds2.lower,
            bounds1.upper + bounds2.upper
        )

    def propagate_size(self, in_size):
        out_size1 = self.bound_network1.propagate_size(in_size)
        out_size2 = self.bound_network2.propagate_size(in_size)

        assert out_size1 == out_size2

        return out_size1

    def bound_parameters(self):
        yield from self.bound_network1.bound_parameters()
        yield from self.bound_network2.bound_parameters()

    def clip_params(self):
        self.bound_network1.clip_params()
        self.bound_network2.clip_params()

    def project_grads(self):
        self.bound_network1.project_grads()
        self.bound_network2.project_grads()


class VectorAdd(nn.Module):
    def forward(self, x):
        assert x.size(-1) % 2 == 0
        half_size = x.size(-1) // 2

        return x[..., :half_size] + x[..., half_size:]


class BoundVectorAdd(BoundModule):
    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        if linear_bounds.lower is None:
            lower = None
        else:
            lower = (torch.cat(2 * [linear_bounds.lower[0]], dim=-1), linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = (torch.cat(2 * [linear_bounds.upper[0]], dim=-1), linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        half_size = bounds.lower.size(-1) // 2

        return IntervalBounds(
            bounds.region,
            bounds.lower[..., :half_size] + bounds.lower[..., half_size:],
            bounds.upper[..., :half_size] + bounds.upper[..., half_size:]
        )

    def propagate_size(self, in_size):
        assert in_size % 2 == 0

        return in_size // 2


class Residual(Add):
    def __init__(self, subnetwork):
        super().__init__(Identity(), subnetwork)


class Sub(nn.Module):
    def __init__(self, network1, network2):
        super().__init__()

        self.network1, self.network2 = network1, network2

    def forward(self, x):
        return self.network1(x) - self.network2(x)


class BoundSub(BoundModule):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.bound_network1 = factory.build(module.network1)
        self.bound_network2 = factory.build(module.network2)

    @property
    def need_relaxation(self):
        return self.bound_network1.need_relaxation or self.bound_network2.need_relaxation

    def clear_relaxation(self):
        self.bound_network1.clear_relaxation()
        self.bound_network2.clear_relaxation()

    def backward_relaxation(self, region):
        if self.bound_network1.need_relaxation:
            return self.bound_network1.backward_relaxation(region)
        else:
            assert self.bound_network2.need_relaxation
            return self.bound_network2.backward_relaxation(region)

    def crown_backward(self, linear_bounds, optimize):
        input_bounds = LinearBounds(
            linear_bounds.region,
            (linear_bounds.lower[0], torch.zeros_like(linear_bounds.lower[1])) if linear_bounds.lower is not None else None,
            (linear_bounds.upper[0], torch.zeros_like(linear_bounds.upper[1])) if linear_bounds.upper is not None else None,
        )
        linear_bounds1 = self.bound_network1.crown_backward(input_bounds, optimize)

        input_bounds = LinearBounds(
            linear_bounds.region,
            (-linear_bounds.lower[0], torch.zeros_like(linear_bounds.lower[1])) if linear_bounds.lower is not None else None,
            (-linear_bounds.upper[0], torch.zeros_like(linear_bounds.upper[1])) if linear_bounds.upper is not None else None,
        )
        linear_bounds2 = self.bound_network2.crown_backward(input_bounds, optimize)

        if linear_bounds.lower is None:
            lower = None
        else:
            lower = (linear_bounds1.lower[0] + linear_bounds2.lower[0], linear_bounds1.lower[1] + linear_bounds2.lower[1] + linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = (linear_bounds1.upper[0] + linear_bounds2.upper[0], linear_bounds1.upper[1] + linear_bounds2.upper[1] + linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        bounds1 = self.bound_network1.ibp_forward(bounds, save_relaxation=save_relaxation, save_input_bounds=save_input_bounds)
        bounds2 = self.bound_network2.ibp_forward(bounds, save_relaxation=save_relaxation, save_input_bounds=save_input_bounds)

        return IntervalBounds(
            bounds.region,
            bounds1.lower - bounds2.upper,
            bounds1.upper - bounds2.lower
        )

    def propagate_size(self, in_size):
        out_size1 = self.bound_network1.propagate_size(in_size)
        out_size2 = self.bound_network2.propagate_size(in_size)

        assert out_size1 == out_size2

        return out_size1

    def bound_parameters(self):
        yield from self.bound_network1.bound_parameters()
        yield from self.bound_network2.bound_parameters()

    def clip_params(self):
        self.bound_network1.clip_params()
        self.bound_network2.clip_params()

    def project_grads(self):
        self.bound_network1.project_grads()
        self.bound_network2.project_grads()


class VectorSub(nn.Module):
    def forward(self, x):
        assert x.size(-1) % 2 == 0
        half_size = x.size(-1) // 2

        return x[..., :half_size] - x[..., half_size:]


class BoundVectorSub(BoundModule):
    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        if linear_bounds.lower is None:
            lower = None
        else:
            lower = (torch.cat([linear_bounds.lower[0], -linear_bounds.lower[0]], dim=-1), linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = (torch.cat([linear_bounds.upper[0], -linear_bounds.upper[0]], dim=-1), linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        half_size = bounds.lower.size(-1) // 2

        return IntervalBounds(
            bounds.region,
            bounds.lower[..., :half_size] - bounds.upper[..., half_size:],
            bounds.upper[..., :half_size] - bounds.lower[..., half_size:]
        )

    def propagate_size(self, in_size):
        assert in_size % 2 == 0

        return in_size // 2


@torch.jit.script
def crown_backward_mul_jit(W_tilde: torch.Tensor, alpha_x: Tuple[torch.Tensor, torch.Tensor], alpha_y: Tuple[torch.Tensor, torch.Tensor], beta: Tuple[torch.Tensor, torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    _delta = torch.where(W_tilde < 0, beta[0].unsqueeze(-2), beta[1].unsqueeze(-2))
    bias = torch.sum(W_tilde * _delta, dim=-1)

    _lambda = torch.where(W_tilde < 0, alpha_x[0].unsqueeze(-2), alpha_x[1].unsqueeze(-2))
    W_tilde_x = W_tilde * _lambda

    _lambda = torch.where(W_tilde < 0, alpha_y[0].unsqueeze(-2), alpha_y[1].unsqueeze(-2))
    W_tilde_y = W_tilde * _lambda

    return W_tilde_x, W_tilde_y, bias


class Mul(nn.Module):
    def __init__(self, network1, network2):
        super().__init__()

        self.network1, self.network2 = network1, network2

    def forward(self, x):
        return self.network1(x) * self.network2(x)


class BoundMul(BoundModule):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.bound_network1 = factory.build(module.network1)
        self.bound_network2 = factory.build(module.network2)

        self.input_bounds = None
        self.kappa_lower, self.kappa_upper = None, None

        self.bounded = False
        self.size = None

    def alpha_beta(self, optimize):
        lower_x, upper_x = self.input_bounds[0].lower, self.input_bounds[0].upper
        lower_y, upper_y = self.input_bounds[1].lower, self.input_bounds[1].upper

        # Lower bound
        first_corner = lower_x, lower_y
        second_corner = upper_x, upper_y

        kappa_lower = self.kappa_lower if optimize else 0.5

        d = first_corner[0] * kappa_lower + second_corner[0] * (1 - kappa_lower), first_corner[1] * kappa_lower + second_corner[1] * (1 - kappa_lower)
        d_prime = d[1], d[0]

        alpha_x_lower = d_prime[0]
        alpha_y_lower = d_prime[1]

        beta_corner = lower_x, upper_y
        beta_lower = beta_corner[0] * beta_corner[1] - d_prime[0] * beta_corner[0] - d_prime[1] * beta_corner[1]

        # Upper bound
        first_corner = lower_x, upper_y
        second_corner = upper_x, lower_y

        kappa_upper = self.kappa_upper if optimize else 0.5

        d = first_corner[0] * kappa_upper + second_corner[0] * (1 - kappa_upper), first_corner[1] * kappa_upper + second_corner[1] * (1 - kappa_upper)
        d_prime = d[1], d[0]

        alpha_x_upper = d_prime[0]
        alpha_y_upper = d_prime[1]

        beta_corner = lower_x, lower_y
        beta_upper = beta_corner[0] * beta_corner[1] - d_prime[0] * beta_corner[0] - d_prime[1] * beta_corner[1]

        return (alpha_x_lower, alpha_x_upper), (alpha_y_lower, alpha_y_upper), (beta_lower, beta_upper)

    def init_kappa(self):
        self.kappa_lower = torch.full_like(self.input_bounds[0].lower, 0.5).requires_grad_()
        self.kappa_upper = torch.full_like(self.input_bounds[0].lower, 0.5).requires_grad_()

    @property
    def need_relaxation(self):
        return self.bound_network1.need_relaxation or self.bound_network2.need_relaxation or not self.bounded

    def set_relaxation(self, linear_bounds):
        interval_bounds = linear_bounds.concretize()
        lower_x, lower_y = interval_bounds.lower.tensor_split(2, dim=-1)
        upper_x, upper_y = interval_bounds.upper.tensor_split(2, dim=-1)

        self.input_bounds = IntervalBounds(linear_bounds.region,
                                           torch.max(lower_x, self.input_bounds[0].lower),
                                           torch.min(upper_x, self.input_bounds[0].upper)),\
                            IntervalBounds(linear_bounds.region,
                                           torch.max(lower_y, self.input_bounds[1].lower),
                                           torch.min(upper_y, self.input_bounds[1].upper))

        self.init_kappa()
        self.bounded = True

    def backward_relaxation(self, region):
        if self.bound_network1.need_relaxation:
            return self.bound_network1.backward_relaxation(region)
        elif self.bound_network2.need_relaxation:
            return self.bound_network2.backward_relaxation(region)
        else:
            assert self.size is not None

            linear_bounds = self.initial_linear_bounds(region, self.size)
            linear_bounds1 = self.bound_network1.crown_backward(linear_bounds, False)
            linear_bounds2 = self.bound_network2.crown_backward(linear_bounds, False)

            lower = (torch.cat([linear_bounds1.lower[0], linear_bounds2.lower[0]], dim=-2), torch.cat([linear_bounds1.lower[1], linear_bounds2.lower[1]], dim=-1))
            upper = (torch.cat([linear_bounds1.upper[0], linear_bounds2.upper[0]], dim=-2), torch.cat([linear_bounds1.upper[1], linear_bounds2.upper[1]], dim=-1))

            return LinearBounds(linear_bounds.region, lower, upper), self

    def clear_relaxation(self):
        self.input_bounds = None
        self.kappa_lower, self.kappa_upper = None, None

        self.bounded = False

    def crown_backward(self, linear_bounds, optimize):
        assert self.bounded

        (alpha_x_lower, alpha_x_upper), (alpha_y_lower, alpha_y_upper), (beta_lower, beta_upper) = self.alpha_beta(optimize)

        # NOTE: The order of alpha and beta are deliberately reverse - this is not a mistake!
        if linear_bounds.lower is None:
            lower = None
        else:
            alpha_x = alpha_x_upper, alpha_x_lower
            alpha_y = alpha_y_upper, alpha_y_lower
            beta = beta_upper, beta_lower

            lower = crown_backward_mul_jit(linear_bounds.lower[0], alpha_x, alpha_y, beta)

        if linear_bounds.upper is None:
            upper = None
        else:
            alpha_x = alpha_x_lower, alpha_x_upper
            alpha_y = alpha_y_lower, alpha_y_upper
            beta = beta_lower, beta_upper

            upper = crown_backward_mul_jit(linear_bounds.upper[0], alpha_x, alpha_y, beta)

        input_bounds1 = LinearBounds(linear_bounds.region,
                                     (lower[0], torch.zeros_like(lower[2])) if lower is not None else None,
                                     (upper[0], torch.zeros_like(upper[2])) if upper is not None else None)
        input_bounds2 = LinearBounds(linear_bounds.region,
                                     (lower[1], torch.zeros_like(lower[2])) if lower is not None else None,
                                     (upper[1], torch.zeros_like(upper[2])) if upper is not None else None)

        linear_bounds1 = self.bound_network1.crown_backward(input_bounds1, optimize)
        linear_bounds2 = self.bound_network2.crown_backward(input_bounds2, optimize)

        if lower is None:
            lower = None
        else:
            lower = (linear_bounds1.lower[0] + linear_bounds2.lower[0], linear_bounds1.lower[1] + linear_bounds2.lower[1] + lower[2] + linear_bounds.lower[1])

        if upper is None:
            upper = None
        else:
            upper = (linear_bounds1.upper[0] + linear_bounds2.upper[0], linear_bounds1.upper[1] + linear_bounds2.upper[1] + upper[2] + linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        bounds1 = self.bound_network1.ibp_forward(bounds, save_relaxation=save_relaxation, save_input_bounds=save_input_bounds)
        bounds2 = self.bound_network2.ibp_forward(bounds, save_relaxation=save_relaxation, save_input_bounds=save_input_bounds)

        if save_relaxation:
            self.input_bounds = bounds1, bounds2
            self.init_kappa()
            self.bounded = True

        if save_input_bounds:
            self.input_bounds = bounds1, bounds2

        combinations = torch.stack([
            bounds1.lower * bounds2.lower,
            bounds1.lower * bounds2.upper,
            bounds1.upper * bounds2.lower,
            bounds1.upper * bounds2.upper
        ], dim=-1)

        return IntervalBounds(
            bounds.region,
            combinations.min(dim=-1).values,
            combinations.max(dim=-1).values
        )

    def propagate_size(self, in_size):
        out_size1 = self.bound_network1.propagate_size(in_size)
        out_size2 = self.bound_network2.propagate_size(in_size)

        assert out_size1 == out_size2

        self.size = out_size1
        return out_size1

    def bound_parameters(self):
        if self.kappa_lower is None or self.kappa_upper is None:
            logger.warning('VectorMul bound not parameterized but expected to')

        yield self.kappa_lower
        yield self.kappa_upper

        yield from self.bound_network1.bound_parameters()
        yield from self.bound_network2.bound_parameters()

    def clip_params(self):
        clip_param_to_range_(self.kappa_lower, (0.0, 1.0))
        clip_param_to_range_(self.kappa_upper, (0.0, 1.0))

        self.bound_network1.clip_params()
        self.bound_network2.clip_params()

    def project_grads(self):
        proj_grad_to_range_(self.kappa_lower, (0.0, 1.0))
        proj_grad_to_range_(self.kappa_upper, (0.0, 1.0))

        self.bound_network1.project_grads()
        self.bound_network2.project_grads()


class Div(Mul):
    def __init__(self, network1, network2):
        super().__init__(network1, nn.Sequential(network2, Reciprocal()))


class VectorMul(nn.Module):
    def forward(self, x):
        assert x.size(-1) % 2 == 0
        half_size = x.size(-1) // 2

        return x[..., :half_size] * x[..., half_size:]


class BoundVectorMul(BoundModule):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.input_bounds = None
        self.kappa_lower, self.kappa_upper = None, None

        self.bounded = False
        self.size = None

    def alpha_beta(self, optimize):
        preactivation = self.input_bounds
        half_size = preactivation.lower.size(-1) // 2

        lower_x, upper_x = preactivation.lower[..., :half_size], preactivation.upper[..., :half_size]
        lower_y, upper_y = preactivation.lower[..., half_size:], preactivation.upper[..., half_size:]

        # Lower bound
        first_corner = lower_x, lower_y
        second_corner = upper_x, upper_y

        kappa_lower = self.kappa_lower if optimize else 0.5

        d = first_corner[0] * kappa_lower + second_corner[0] * (1 - kappa_lower), first_corner[1] * kappa_lower + second_corner[1] * (1 - kappa_lower)
        d_prime = d[1], d[0]

        alpha_x_lower = d_prime[0]
        alpha_y_lower = d_prime[1]

        beta_corner = lower_x, upper_y
        beta_lower = beta_corner[0] * beta_corner[1] - d_prime[0] * beta_corner[0] - d_prime[1] * beta_corner[1]

        # Upper bound
        first_corner = lower_x, upper_y
        second_corner = upper_x, lower_y

        kappa_upper = self.kappa_upper if optimize else 0.5

        d = first_corner[0] * kappa_upper + second_corner[0] * (1 - kappa_upper), first_corner[1] * kappa_upper + second_corner[1] * (1 - kappa_upper)
        d_prime = d[1], d[0]

        alpha_x_upper = d_prime[0]
        alpha_y_upper = d_prime[1]

        beta_corner = lower_x, lower_y
        beta_upper = beta_corner[0] * beta_corner[1] - d_prime[0] * beta_corner[0] - d_prime[1] * beta_corner[1]

        return (alpha_x_lower, alpha_x_upper), (alpha_y_lower, alpha_y_upper), (beta_lower, beta_upper)

    def init_kappa(self):
        half_size = self.input_bounds.lower.size(-1) // 2
        lower_x = self.input_bounds.lower[..., :half_size]

        self.kappa_lower = torch.full_like(lower_x, 0.5).requires_grad_()
        self.kappa_upper = torch.full_like(lower_x, 0.5).requires_grad_()

    @property
    def need_relaxation(self):
        return not self.bounded

    def set_relaxation(self, linear_bounds):
        interval_bounds = linear_bounds.concretize()
        self.input_bounds = IntervalBounds(
            linear_bounds.region,
            torch.max(interval_bounds.lower, self.input_bounds.lower),
            torch.min(interval_bounds.upper, self.input_bounds.upper)
        )

        self.init_kappa()
        self.bounded = True

    def backward_relaxation(self, region):
        assert self.size is not None

        linear_bounds = self.initial_linear_bounds(region, self.size)
        return linear_bounds, self

    def clear_relaxation(self):
        self.input_bounds = None
        self.kappa_lower, self.kappa_upper = None, None

        self.bounded = False

    def crown_backward(self, linear_bounds, optimize):
        assert self.bounded

        (alpha_x_lower, alpha_x_upper), (alpha_y_lower, alpha_y_upper), (beta_lower, beta_upper) = self.alpha_beta(optimize)

        # NOTE: The order of alpha and beta are deliberately reverse - this is not a mistake!
        if linear_bounds.lower is None:
            lower = None
        else:
            alpha_x = alpha_x_upper, alpha_x_lower
            alpha_y = alpha_y_upper, alpha_y_lower
            beta = beta_upper, beta_lower

            lower = crown_backward_mul_jit(linear_bounds.lower[0], alpha_x, alpha_y, beta)
            lower = (torch.cat(lower[:2], dim=-1), lower[2] + linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            alpha_x = alpha_x_lower, alpha_x_upper
            alpha_y = alpha_y_lower, alpha_y_upper
            beta = beta_lower, beta_upper

            upper = crown_backward_mul_jit(linear_bounds.upper[0], alpha_x, alpha_y, beta)
            upper = (torch.cat(upper[:2], dim=-1), upper[2] + linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        if save_relaxation:
            self.input_bounds = bounds
            self.init_kappa()
            self.bounded = True

        if save_input_bounds:
            self.input_bounds = bounds

        half_size = bounds.lower.size(-1) // 2

        lower_x, upper_x = bounds.lower[..., :half_size], bounds.upper[..., :half_size]
        lower_y, upper_y = bounds.lower[..., half_size:], bounds.upper[..., half_size:]

        combinations = torch.stack([
            lower_x * lower_y,
            lower_x * upper_y,
            upper_x * lower_y,
            upper_x * upper_y
        ], dim=-1)

        return IntervalBounds(
            bounds.region,
            combinations.min(dim=-1).values,
            combinations.max(dim=-1).values
        )

    def propagate_size(self, in_size):
        assert in_size % 2 == 0

        self.size = in_size
        return in_size // 2

    def bound_parameters(self):
        if self.kappa_lower is None or self.kappa_upper is None:
            logger.warning('VectorMul bound not parameterized but expected to')

        yield self.kappa_lower
        yield self.kappa_upper

    def clip_params(self):
        clip_param_to_range_(self.kappa_lower, (0.0, 1.0))
        clip_param_to_range_(self.kappa_upper, (0.0, 1.0))

    def project_grads(self):
        proj_grad_to_range_(self.kappa_lower, (0.0, 1.0))
        proj_grad_to_range_(self.kappa_upper, (0.0, 1.0))



import abc
import logging
from typing import Tuple

import numpy as np
import torch
from torch import nn

from . import IntervalBounds
from .saturation import Clamp
from .bivariate import Div, VectorSub
from .reshape import Flip
from .activation import BoundSigmoid, BoundActivation, assert_bound_order, bisection
from .linear import ElementWiseLinear
from .util import proj_grad_to_range_, clip_param_to_range_

logger = logging.getLogger(__name__)


class Erf(nn.Module):
    def forward(self, x):
        return torch.special.erf(x)


class BoundErf(BoundSigmoid):
    def derivative(self, x):
        return (2 / np.sqrt(np.pi)) * torch.exp(-x.pow(2))


class StandardNormalPDF(nn.Module):
    # While this class could be made as a sequential with pow(2), linear, exponential, linear,
    # by explicitly bounding a bell curve, the bounds will be tighter. Furthermore, BoundBellCurve
    # will be useful for other bell curve-shaped functions (in particular other distributions).

    def forward(self, x):
        return (1 / np.sqrt(2 * np.pi)) * torch.exp(-0.5 * x.pow(2))


def bell_curve_regimes(lower: torch.Tensor, upper: torch.Tensor, top_point) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    zero_width = torch.isclose(lower, upper, rtol=0.0, atol=1e-8) & torch.full_like(top_point, True, dtype=torch.bool)
    l = (~zero_width) & (upper <= top_point)
    u = (~zero_width) & (top_point <= lower)
    lu = (~zero_width) & (lower < top_point) & (top_point < upper)

    return zero_width, l, u, lu


class BoundBellCurve(BoundActivation, abc.ABC):
    midpoint = 0.0
    # - For some bell curves (non-Gaussian), the inflection points are unknown, but we do know that there are two and
    # we may compute upper and lower bounds for each inflection point (e.g. using bisection).
    # - Also note that inflection points may be tensors to allow data parallel computation of linear bounds for
    # multiple different bell curves.
    lower_inflection = (-1.0, -1.0)
    upper_inflection = (1.0, 1.0)

    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def clear_relaxation(self):
        super().clear_relaxation()
        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    @abc.abstractmethod
    def derivative(self, x, mask=None):
        raise NotImplementedError()

    def func(self, x, mask=None):
        return self(x)

    @assert_bound_order
    def alpha_beta(self, preactivation):
        lower, upper = preactivation.lower, preactivation.upper

        zero_width, l, u, lu = bell_curve_regimes(lower, upper, self.midpoint)
        ones = torch.ones_like(zero_width, dtype=lower.dtype)
        lower, upper = ones * lower, ones * upper

        self.alpha_lower, self.beta_lower = torch.zeros_like(lower), torch.zeros_like(lower)
        self.alpha_upper, self.beta_upper = torch.zeros_like(lower), torch.zeros_like(lower)

        lower_act, upper_act = self(lower), self(upper)
        lower_prime, upper_prime = self.derivative(lower), self.derivative(upper)
        slope = (upper_act - lower_act) / (upper - lower)
        lower_inflection = self.lower_inflection[0] * ones, self.lower_inflection[1] * ones
        upper_inflection = self.upper_inflection[0] * ones, self.upper_inflection[1] * ones
        mid = self.midpoint * ones

        # Use upper and lower in the bias to account for a small numerical difference between lower and upper
        # which ought to be negligible, but may still be present due to torch.isclose.
        self.alpha_lower[zero_width], self.beta_lower[zero_width] = 0, torch.min(lower_act[zero_width], upper_act[zero_width])
        self.alpha_upper[zero_width], self.beta_upper[zero_width] = 0, torch.max(lower_act[zero_width], upper_act[zero_width])

        def add_linear(alpha, beta, mask, a, x, y):
            alpha[mask] = a
            beta[mask] = y - a * x

        ###############
        # Lower bound #
        ###############
        lower_is_closer = ((lower - self.midpoint).abs() <= (upper - self.midpoint).abs())
        optimize_upper = (lu & lower_is_closer) | u
        optimize_lower = (lu & (~lower_is_closer)) | l

        # Fix lower input
        # - If we can take the direct slope
        direct_cond = ((lower < self.lower_inflection[1]) & (lower_prime >= slope)) | (lower >= self.lower_inflection[1])
        direct = optimize_lower & direct_cond
        add_linear(self.alpha_lower, self.beta_lower, mask=direct, a=slope[direct], x=upper[direct], y=upper_act[direct])

        # - Else
        indirect = optimize_lower & (~direct_cond)
        # - Bound right input such that the tangent lines are true upper bounds
        right = self.lower_left_tangent_point(indirect, lower, upper, lower_inflection)

        d = (lower[indirect] + right) / 2
        add_linear(self.alpha_lower, self.beta_lower, mask=indirect, a=self.derivative(d, mask=indirect), x=d, y=self.func(d, mask=indirect))

        # Allow parameterization
        # Save mask
        unstable_lower = [indirect]
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        unstable_d_lower = [d.detach().clone().requires_grad_()]
        # Save ranges to clip (aka. PGD)
        unstable_range_lower = [(lower[indirect], right)]

        # Fix upper input
        # - If we can take the direct slope
        direct_cond = ((upper > self.upper_inflection[0]) & (upper_prime <= slope)) | (upper <= self.upper_inflection[0])
        direct = optimize_upper & direct_cond
        add_linear(self.alpha_lower, self.beta_lower, mask=direct, a=slope[direct], x=lower[direct], y=lower_act[direct])

        # - Else
        indirect = optimize_upper & (~direct_cond)
        # - Bound left input such that the tangent lines are true upper bounds
        left = self.lower_right_tangent_point(indirect, lower, upper, upper_inflection)

        d = (left + upper[indirect]) / 2
        add_linear(self.alpha_lower, self.beta_lower, mask=indirect, a=self.derivative(d, mask=indirect), x=d, y=self.func(d, mask=indirect))

        # Allow parameterization
        # Save mask
        unstable_lower.append(indirect)
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        unstable_d_lower.append(d.detach().clone().requires_grad_())
        # Save ranges to clip (aka. PGD)
        unstable_range_lower.append((left, upper[indirect]))

        self.unstable_lower = unstable_lower
        self.unstable_d_lower = unstable_d_lower
        self.unstable_range_lower = unstable_range_lower

        ################
        # Lower regime #
        ################
        # Upper bound
        # - If we can take the direct slope
        direct_cond = ((upper > self.lower_inflection[0]) & (upper_prime >= slope)) | (upper <= self.lower_inflection[0])
        direct = l & direct_cond
        add_linear(self.alpha_upper, self.beta_upper, mask=direct, a=slope[direct], x=upper[direct], y=upper_act[direct])

        # - Else
        indirect = l & (~direct_cond)
        # - Bound left input such that the tangent lines are true upper bounds
        left = self.upper_right_tangent_point(indirect, lower, upper, mid, lower_inflection)

        d = (left + upper[indirect]) / 2
        add_linear(self.alpha_upper, self.beta_upper, mask=indirect, a=self.derivative(d, mask=indirect), x=d, y=self.func(d, mask=indirect))

        # Allow parameterization
        # Save mask
        unstable_upper = [indirect]
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        unstable_d_upper = [d.detach().clone().requires_grad_()]
        # Save ranges to clip (aka. PGD)
        unstable_range_upper = [(left, upper[indirect])]

        ################
        # Upper regime #
        ################
        # Upper bound
        # - If we can take the direct slope
        direct_cond = ((lower < self.upper_inflection[1]) & (lower_prime <= slope)) | (lower >= self.upper_inflection[1])
        direct = u & direct_cond
        add_linear(self.alpha_upper, self.beta_upper, mask=direct, a=slope[direct], x=lower[direct], y=lower_act[direct])

        # - Else
        indirect = u & (~direct_cond)
        # - Bound right input such that the tangent lines are true upper bounds
        right = self.upper_left_tangent_point(indirect, lower, upper, mid, upper_inflection)

        d = (lower[indirect] + right) / 2
        add_linear(self.alpha_upper, self.beta_upper, mask=indirect, a=self.derivative(d, mask=indirect), x=d, y=self.func(d, mask=indirect))

        # Allow parameterization
        # Save mask
        unstable_upper.append(indirect)
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        unstable_d_upper.append(d.detach().clone().requires_grad_())
        # Save ranges to clip (aka. PGD)
        unstable_range_upper.append((lower[indirect], right))

        #####################
        # Crossing midpoint #
        #####################
        # Upper bound
        # - Bound left and right inputs such that the tangent lines are true upper bounds
        left = self.upper_right_tangent_point(lu, lower, upper, mid, lower_inflection)
        right = self.upper_left_tangent_point(lu, lower, upper, mid, upper_inflection)

        d = (left + right) / 2
        add_linear(self.alpha_upper, self.beta_upper, mask=lu, a=self.derivative(d, mask=lu), x=d, y=self.func(d, mask=lu))

        # Allow parameterization
        # Save mask
        unstable_upper.append(lu)
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        unstable_d_upper.append(d.detach().clone().requires_grad_())
        # Save ranges to clip (aka. PGD)
        unstable_range_upper.append((left, right))

        self.unstable_upper = unstable_upper
        self.unstable_d_upper = unstable_d_upper
        self.unstable_range_upper = unstable_range_upper

    def upper_right_tangent_point(self, mask, lower, upper, mid, lower_inflection):
        over_inflection = lower >= lower_inflection[1]
        direct, indirect = mask & over_inflection, mask & (~over_inflection)
        over_inflection = over_inflection[mask]

        d_out = torch.zeros_like(lower[mask])
        d_out[over_inflection] = lower[direct]

        lower, upper = lower[indirect], upper[indirect]
        lower_act = self.func(lower, mask=indirect)

        def f(d: torch.Tensor) -> torch.Tensor:
            a_slope = (self.func(d, mask=indirect) - lower_act) / (d - lower)
            a_derivative = self.derivative(d, mask=indirect)
            return a_slope - a_derivative

        # Bisection will return left and right bounds for d s.t. f(d) is zero
        # Derivative of right bound will over-approximate the slope - hence a true bound
        bisection_lower = lower_inflection[1][indirect]
        bisection_upper = mid[indirect]
        _, d = bisection(bisection_lower, bisection_upper, f)
        d_out[~over_inflection] = d

        return d_out

    def upper_left_tangent_point(self, mask, lower, upper, mid, upper_inflection):
        under_inflection = upper <= upper_inflection[0]
        direct, indirect = mask & under_inflection, mask & (~under_inflection)
        under_inflection = under_inflection[mask]

        d_out = torch.zeros_like(lower[mask])
        d_out[under_inflection] = upper[direct]

        lower, upper = lower[indirect], upper[indirect]
        upper_act = self.func(upper, mask=indirect)

        def f(d: torch.Tensor) -> torch.Tensor:
            a_slope = (upper_act - self.func(d, mask=indirect)) / (upper - d)
            a_derivative = self.derivative(d, mask=indirect)
            return a_slope - a_derivative

        # Bisection will return left and right bounds for d s.t. f(d) is zero
        # Derivative of left bound will over-approximate the slope - hence a true bound
        bisection_lower = mid[indirect]
        bisection_upper = upper_inflection[0][indirect]
        d, _ = bisection(bisection_lower, bisection_upper, f)
        d_out[~under_inflection] = d

        return d_out

    def lower_left_tangent_point(self, mask, lower, upper, lower_inflection):
        under_inflection = upper <= lower_inflection[0]
        direct, indirect = mask & under_inflection, mask & (~under_inflection)
        under_inflection = under_inflection[mask]

        d_out = torch.zeros_like(lower[mask])
        d_out[under_inflection] = upper[direct]

        lower, upper = lower[indirect], upper[indirect]
        upper_act = self.func(upper, mask=indirect)

        def f(d: torch.Tensor) -> torch.Tensor:
            a_slope = (upper_act - self.func(d, mask=indirect)) / (upper - d)
            a_derivative = self.derivative(d, mask=indirect)
            return a_derivative - a_slope

        # Bisection will return left and right bounds for d s.t. f(d) is zero
        # Derivative of left bound will over-approximate the slope - hence a true bound
        bisection_lower = lower
        bisection_upper = lower_inflection[0][indirect]
        d, _ = bisection(bisection_lower, bisection_upper, f)
        d_out[~under_inflection] = d

        return d_out

    def lower_right_tangent_point(self, mask, lower, upper, upper_inflection):
        over_inflection = lower >= upper_inflection[1]
        direct, indirect = mask & over_inflection, mask & (~over_inflection)
        over_inflection = over_inflection[mask]

        d_out = torch.zeros_like(lower[mask])
        d_out[over_inflection] = lower[direct]

        lower, upper = lower[indirect], upper[indirect]
        lower_act = self.func(lower, mask=indirect)

        def f(d: torch.Tensor) -> torch.Tensor:
            a_slope = (self.func(d, mask=indirect) - lower_act) / (d - lower)
            a_derivative = self.derivative(d, mask=indirect)
            return a_derivative - a_slope

        # Bisection will return left and right bounds for d s.t. f(d) is zero
        # Derivative of right bound will over-approximate the slope - hence a true bound
        bisection_lower = upper_inflection[1][indirect]
        bisection_upper = upper
        _, d = bisection(bisection_lower, bisection_upper, f)
        d_out[~over_inflection] = d

        return d_out

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        if save_relaxation:
            self.alpha_beta(preactivation=bounds)
            self.bounded = True

        if save_input_bounds:
            self.input_bounds = bounds

        zero_width, l, u, lu = bell_curve_regimes(bounds.lower, bounds.upper, self.midpoint)

        lower = torch.zeros_like(zero_width, dtype=bounds.lower.dtype)
        upper = torch.zeros_like(zero_width, dtype=bounds.lower.dtype)

        lower_act = self(bounds.lower)
        upper_act = self(bounds.upper)

        lower[zero_width] = torch.min(lower_act[zero_width], upper_act[zero_width])
        upper[zero_width] = torch.max(lower_act[zero_width], upper_act[zero_width])

        lower[l] = lower_act[l]
        upper[l] = upper_act[l]

        lower[u] = upper_act[u]
        upper[u] = lower_act[u]

        lower[lu] = torch.min(lower_act[lu], upper_act[lu])
        upper[lu] = (self(torch.as_tensor(self.midpoint, device=lower.device, dtype=upper.dtype)) * torch.ones_like(lower_act))[lu]

        return IntervalBounds(bounds.region, lower, upper)

    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Bell curve bound not parameterized but expected to')

        # Use implicit parameterization (i.e. store d [point where touching the curve], and not alpha)
        def add_linear(alpha, beta, mask, x):
            a = self.derivative(x, mask=mask)
            y = self.func(x, mask=mask)

            alpha[mask] = a
            beta[mask] = y - a * x

        add_linear(alpha_lower, beta_lower, mask=self.unstable_lower[0], x=self.unstable_d_lower[0])
        add_linear(alpha_lower, beta_lower, mask=self.unstable_lower[1], x=self.unstable_d_lower[1])

        add_linear(alpha_upper, beta_upper, mask=self.unstable_upper[0], x=self.unstable_d_upper[0])
        add_linear(alpha_upper, beta_upper, mask=self.unstable_upper[1], x=self.unstable_d_upper[1])
        add_linear(alpha_upper, beta_upper, mask=self.unstable_upper[2], x=self.unstable_d_upper[2])

        return alpha_lower, alpha_upper, beta_lower, beta_upper

    def bound_parameters(self):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Bell curve bound not parameterized but expected to')

        yield from self.unstable_d_lower
        yield from self.unstable_d_upper

    def clip_params(self):
        clip_param_to_range_(self.unstable_d_lower[0], self.unstable_range_lower[0])
        clip_param_to_range_(self.unstable_d_lower[1], self.unstable_range_lower[1])

        clip_param_to_range_(self.unstable_d_upper[0], self.unstable_range_upper[0])
        clip_param_to_range_(self.unstable_d_upper[1], self.unstable_range_upper[1])
        clip_param_to_range_(self.unstable_d_upper[2], self.unstable_range_upper[2])

    def project_grads(self):
        proj_grad_to_range_(self.unstable_d_lower[0], self.unstable_range_lower[0])
        proj_grad_to_range_(self.unstable_d_lower[1], self.unstable_range_lower[1])

        proj_grad_to_range_(self.unstable_d_upper[0], self.unstable_range_upper[0])
        proj_grad_to_range_(self.unstable_d_upper[1], self.unstable_range_upper[1])
        proj_grad_to_range_(self.unstable_d_upper[2], self.unstable_range_upper[2])


class BoundStandardNormalPDF(BoundBellCurve):
    midpoint = 0.0
    lower_infliction = (-1.0, -1.0)
    upper_infliction = (1.0, 1.0)

    def derivative(self, x):
        return (1 / np.sqrt(2 * np.pi)) * (-torch.exp(-0.5 * x.pow(2)) * x)


class StandardNormalCDF(nn.Sequential):
    def __init__(self):
        super().__init__(
            ElementWiseLinear(1 / np.sqrt(2)),
            Erf(),
            ElementWiseLinear(0.5, 0.5)
        )


class NormalPDF(nn.Sequential):
    def __init__(self, loc, scale):
        inv_scale = 1 / scale

        super().__init__(
            ElementWiseLinear(inv_scale, -loc * inv_scale),
            StandardNormalPDF(),
            ElementWiseLinear(inv_scale)
        )


class NormalCDF(nn.Sequential):
    def __init__(self, loc, scale):
        inv_scale = 1 / (scale * np.sqrt(2))

        super().__init__(
            ElementWiseLinear(inv_scale, -loc * inv_scale),
            Erf(),
            ElementWiseLinear(0.5, 0.5)
        )


class TruncatedGaussianTwoSidedExpectation(nn.Sequential):
    def __init__(self, loc, scale, epsilon=1e-8):
        inv_scale = 1 / scale

        super().__init__(
            ElementWiseLinear(inv_scale, -loc * inv_scale),
            Div(
                nn.Sequential(StandardNormalPDF(), VectorSub()),
                nn.Sequential(Flip(), StandardNormalCDF(), VectorSub(), Clamp(min=epsilon)),
            ),
            ElementWiseLinear(scale, loc)
        )


class TruncatedGaussianLowerTailExpectation(nn.Sequential):
    def __init__(self, loc, scale):
        inv_scale = 1 / scale

        super().__init__(
            ElementWiseLinear(inv_scale, -loc * inv_scale),
            Div(
                StandardNormalPDF(),
                nn.Sequential(StandardNormalCDF(), ElementWiseLinear(-1.0, 1.0)),
            ),
            ElementWiseLinear(scale, loc)
        )


class TruncatedGaussianUpperTailExpectation(nn.Sequential):
    def __init__(self, loc, scale):
        inv_scale = 1 / scale

        super().__init__(
            ElementWiseLinear(inv_scale, -loc * inv_scale),
            Div(
                StandardNormalPDF(),
                nn.Sequential(StandardNormalCDF())
            ),
            ElementWiseLinear(-scale, loc)
        )



import logging
from typing import Tuple, List

import torch
from torch import nn

from .parallel import Parallel
from .reshape import Select
from .bivariate import Mul
from .activation import assert_bound_order, regimes, bisection, BoundActivation
from .bounds import IntervalBounds, LinearBounds
from .general import BoundModule
from .util import clip_param_to_range_, proj_grad_to_range_

logger = logging.getLogger(__name__)


class Pow(nn.Module):
    def __init__(self, power):
        super().__init__()

        # assert power >= 1, 'Pow only supports integer powers'
        # assert all([isinstance(power, int) and power >= 2 for power in self.powers]), 'Univariate monomial only supports integer powers'

        self.power = power

    def forward(self, x):
        return x.pow(self.power)


class BoundPow(BoundActivation):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def clear_relaxation(self):
        super().clear_relaxation()
        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    @assert_bound_order
    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        if save_relaxation:
            self.alpha_beta(preactivation=bounds)
            self.bounded = True

        if save_input_bounds:
            self.input_bounds = bounds

        lower, upper = bounds.lower, bounds.upper
        lower_act, upper_act = self(lower), self(upper)

        lower_out, upper_out = torch.zeros_like(lower), torch.zeros_like(upper)

        # Even powers
        even = (torch.as_tensor(self.module.power, device=lower.device) % 2) == 0
        crossing = (lower < 0) & (upper > 0)

        crossing, not_crossing = (crossing & even), ((~crossing) & even)

        lower_out[crossing] = torch.zeros_like(lower[crossing])
        lower_out[not_crossing] = torch.min(lower_act[not_crossing], upper_act[not_crossing])

        upper_out[..., even] = torch.max(lower_act[..., even], upper_act[..., even])

        # Odd powers
        lower_out[..., ~even] = lower_act[..., ~even]
        upper_out[..., ~even] = upper_act[..., ~even]

        return IntervalBounds(bounds.region, lower_out, upper_out)

    def func(self, x, power=None):
        if power is None:
            power = torch.as_tensor(self.module.power, device=x.device)

        x = x.pow(power)
        return x

    def derivative(self, x, power=None):
        if power is None:
            power = torch.as_tensor(self.module.power, device=x.device)

        x = power * x.pow(power - 1)
        return x

    @assert_bound_order
    def alpha_beta(self, preactivation):
        lower, upper = preactivation.lower, preactivation.upper
        zero_width, n, p, np = regimes(lower, upper)
        power = torch.as_tensor(self.module.power, device=lower.device)
        even = power % 2 == 0
        odd = ~even

        self.alpha_lower, self.beta_lower = torch.zeros_like(lower), torch.zeros_like(lower)
        self.alpha_upper, self.beta_upper = torch.zeros_like(lower), torch.zeros_like(lower)

        lower_act, upper_act = self.func(lower), self.func(upper)
        lower_prime, upper_prime = self.derivative(lower), self.derivative(upper)

        d = (lower + upper) * 0.5  # Let d be the midpoint of the two bounds
        d_act = self.func(d)
        d_prime = self.derivative(d)

        slope = (upper_act - lower_act) / (upper - lower)

        def add_linear(alpha, beta, mask, a, x, y, a_mask=True):
            if a_mask:
                a = a[mask]

            alpha[mask] = a
            beta[mask] = y[mask] - a * x[mask]

        # Use upper and lower in the bias to account for a small numerical difference between lower and upper
        # which ought to be negligible, but may still be present due to torch.isclose.
        self.alpha_lower[zero_width], self.beta_lower[zero_width] = 0, torch.min(lower_act[zero_width], upper_act[zero_width])
        self.alpha_upper[zero_width], self.beta_upper[zero_width] = 0, torch.max(lower_act[zero_width], upper_act[zero_width])

        ########
        # Even #
        ########
        all_even = (n | p | np) & even

        add_linear(self.alpha_lower, self.beta_lower, mask=all_even, a=d_prime, x=d, y=d_act)
        add_linear(self.alpha_upper, self.beta_upper, mask=all_even, a=slope, x=lower, y=lower_act)

        # Allow parameterization
        # Save mask
        unstable_lower = [all_even]
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        unstable_d_lower = [d[all_even].detach().clone().requires_grad_()]
        # Save ranges to clip (aka. PGD)
        unstable_range_lower = [(lower[all_even], upper[all_even])]

        #########################
        # Odd - negative regime #
        #########################
        n_odd = n & odd

        add_linear(self.alpha_lower, self.beta_lower, mask=n_odd, a=slope, x=lower, y=lower_act)
        add_linear(self.alpha_upper, self.beta_upper, mask=n_odd, a=d_prime, x=d, y=d_act)

        # Allow parameterization
        # Save mask
        self.unstable_upper = n_odd
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        self.unstable_d_upper = d[n_odd].detach().clone().requires_grad_()
        # Save ranges to clip (aka. PGD)
        self.unstable_range_upper = lower[n_odd], upper[n_odd]

        #########################
        # Odd - positive regime #
        #########################
        p_odd = p & odd

        add_linear(self.alpha_lower, self.beta_lower, mask=p_odd, a=d_prime, x=d, y=d_act)
        add_linear(self.alpha_upper, self.beta_upper, mask=p_odd, a=slope, x=lower, y=lower_act)

        # Allow parameterization
        # Save mask
        unstable_lower.append(p_odd)
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        unstable_d_lower.append(d[p_odd].detach().clone().requires_grad_())
        # Save ranges to clip (aka. PGD)
        unstable_range_lower.append((lower[p_odd], upper[p_odd]))

        self.unstable_lower = unstable_lower
        self.unstable_d_lower = unstable_d_lower
        self.unstable_range_lower = unstable_range_lower

        #######################
        # Odd - crossing zero #
        #######################
        np_odd = np & odd
        power = power.view(*[1 for _ in range(np.dim() - 1)], -1).expand_as(np_odd)

        # Upper bound #
        # If tangent to lower is below upper, then take direct slope between lower and upper
        direct = np_odd & (slope >= lower_prime)
        add_linear(self.alpha_upper, self.beta_upper, mask=direct, a=slope, x=upper, y=upper_act)

        # Else use bisection to find upper bound on slope.
        implicit = np_odd & (slope < lower_prime)

        if torch.any(implicit):
            implicit_power = power[implicit]
            implicit_lower, implicit_upper = lower[implicit], upper[implicit]
            implicit_upper_act = self.func(implicit_upper, implicit_power)

            def f_upper(d: torch.Tensor) -> torch.Tensor:
                a_slope = (implicit_upper_act - self.func(d, implicit_power)) / (implicit_upper - d)
                a_derivative = self.derivative(d, implicit_power)
                return a_slope - a_derivative

            # Bisection will return left and right bounds for d s.t. f_upper(d) is zero
            # Derivative of left bound will over-approximate the slope - hence a true bound
            d_upper, _ = bisection(implicit_lower, torch.zeros_like(implicit_lower), f_upper)
            # Slope has to attach to (lower, sigma(lower))
            add_linear(self.alpha_upper, self.beta_upper, mask=implicit, a=self.derivative(d_upper, implicit_power), x=upper, y=upper_act, a_mask=False)

        # Lower bound #
        # If tangent to upper is above lower, then take direct slope between lower and upper
        direct = np_odd & (slope >= upper_prime)
        add_linear(self.alpha_lower, self.beta_lower, mask=direct, a=slope, x=lower, y=lower_act)

        # Else use bisection to find upper bound on slope.
        implicit = np_odd & (slope < upper_prime)

        if torch.any(implicit):
            implicit_power = power[implicit]
            implicit_lower, implicit_upper = lower[implicit], upper[implicit]
            implicit_lower_act = self.func(implicit_lower, implicit_power)

            def f_lower(d: torch.Tensor) -> torch.Tensor:
                a_slope = (self.func(d, implicit_power) - implicit_lower_act) / (d - implicit_lower)
                a_derivative = self.derivative(d, implicit_power)
                return a_derivative - a_slope

            # Bisection will return left and right bounds for d s.t. f_lower(d) is zero
            # Derivative of right bound will over-approximate the slope - hence a true bound
            _, d_lower = bisection(torch.zeros_like(implicit_upper), implicit_upper, f_lower)
            # Slope has to attach to (upper, sigma(upper))
            add_linear(self.alpha_lower, self.beta_lower, mask=implicit, a=self.derivative(d_lower, implicit_power), x=lower, y=lower_act, a_mask=False)

    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Polynomial bound not parameterized but expected to')

        power = torch.as_tensor(self.module.power, device=alpha_lower.device).view(*[1 for _ in range(alpha_lower.dim() - 1)], -1).expand_as(alpha_lower)

        # Use implicit parameterization (i.e. store d [point where touching the curve], and not alpha)
        def add_linear(alpha, beta, mask, x):
            p = power[mask]

            a = self.derivative(x, p)
            y = self.func(x, p)

            alpha[mask] = a
            beta[mask] = y - a * x

        add_linear(alpha_lower, beta_lower, mask=self.unstable_lower[0], x=self.unstable_d_lower[0])
        add_linear(alpha_lower, beta_lower, mask=self.unstable_lower[1], x=self.unstable_d_lower[1])
        add_linear(alpha_upper, beta_upper, mask=self.unstable_upper, x=self.unstable_d_upper)

        return alpha_lower, alpha_upper, beta_lower, beta_upper

    def bound_parameters(self):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Polynomial bound not parameterized but expected to')

        yield from self.unstable_d_lower
        yield self.unstable_d_upper

    def clip_params(self):
        clip_param_to_range_(self.unstable_d_lower[0], self.unstable_range_lower[0])
        clip_param_to_range_(self.unstable_d_lower[1], self.unstable_range_lower[1])
        clip_param_to_range_(self.unstable_d_upper, self.unstable_range_upper)

    def project_grads(self):
        proj_grad_to_range_(self.unstable_d_lower[0], self.unstable_range_lower[0])
        proj_grad_to_range_(self.unstable_d_lower[1], self.unstable_range_lower[1])
        proj_grad_to_range_(self.unstable_d_upper, self.unstable_range_upper)


class UnivariateMonomial(nn.Sequential):
    def __init__(self, powers):
        # Assume powers are of the structure [(index, power)] and this will be the output order too
        indices = [index for index, _ in powers]
        powers = [powers for _, powers in powers]

        super().__init__(
            Select(indices),
            Pow(powers)
        )


class MultivariateMonomial(nn.Sequential):
    def __init__(self, monomials):
        super().__init__()

        monomials = [self.coalesce_factors(monomial) for monomial in monomials]

        # Assume powers are of the structure [[(index, power)]]
        linear_terms = list(set([index for monomial in monomials for index, power in monomial if power == 1]))
        factors = list(set([(index, power) for monomial in monomials for index, power in monomial if power >= 2]))
        monomial_indices = []

        for monomial in monomials:
            monomial_index = []
            for index, power in monomial:
                if power == 1:
                    monomial_index.append(linear_terms.index(index))
                else:
                    assert isinstance(power, int) and power >= 2, 'Multivariate monomial only supports integer powers'

                    monomial_index.append(len(linear_terms) + factors.index((index, power)))

            monomial_indices.append(monomial_index)

        super().__init__(
            Parallel(Select(linear_terms), UnivariateMonomial(factors)),
            Parallel(*[self.construct_mul(monomial_index) for monomial_index in monomial_indices])
        )

    def construct_mul(self, monomial_index):
        module = Select(monomial_index[0])

        for index in monomial_index[1:]:
            module = Mul(module, Select(index))

        return module

    def coalesce_factors(self, monomial):
        powers = {}

        for index, power in monomial:
            powers[index] = powers.get(index, 0) + power

        return list(powers.items())


class Polynomials(nn.Module):
    pass

    # Combine Linear and MultivariateMonomial to form Polynomials



from typing import Optional, Tuple, Union

import torch
from torch import nn

from .activation import assert_bound_order
from .general import BoundModule
from .bounds import LinearBounds, IntervalBounds


@torch.jit.script
def crown_backward_linear_jit(weight: torch.Tensor, bias: Optional[torch.Tensor], W_tilde: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    if bias is None:
        bias_acc = torch.tensor(0.0, device=W_tilde.device, dtype=W_tilde.dtype)
    else:
        bias = bias.unsqueeze(-2).to(W_tilde.dtype)
        bias_acc = bias.matmul(W_tilde.transpose(-1, -2)).squeeze(-2)

    weight = weight.to(W_tilde.dtype)
    if weight.dim() == 2:
        W_tilde = W_tilde.matmul(weight)
    else:
        if W_tilde.dim() == 3:
            W_tilde = W_tilde.unsqueeze(0)

        W_tilde = W_tilde.matmul(weight.unsqueeze(1))

    return W_tilde, bias_acc


@torch.jit.script
def ibp_forward_linear_jit(weight: torch.Tensor, bias: Optional[torch.Tensor], center: torch.Tensor, diff: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    center, diff = center.unsqueeze(-2), diff.unsqueeze(-2)
    device, dtype = center.device, center.dtype

    weight = weight.transpose(-1, -2).to(dtype)

    w_mid = center.matmul(weight) + (bias.to(dtype).unsqueeze(-2) if bias is not None else torch.tensor(0.0, device=device, dtype=dtype))
    w_diff = diff.matmul(weight.abs())

    lower = w_mid - w_diff
    lower = lower.squeeze(-2)

    upper = w_mid + w_diff
    upper = upper.squeeze(-2)

    return lower, upper


class BoundLinear(BoundModule):

    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        if linear_bounds.lower is None:
            lower = None
        else:
            lower = crown_backward_linear_jit(self.module.weight, self.module.bias, linear_bounds.lower[0])
            lower = (lower[0], lower[1] + linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = crown_backward_linear_jit(self.module.weight, self.module.bias, linear_bounds.upper[0])
            upper = (upper[0], upper[1] + linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    @assert_bound_order
    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        center, diff = bounds.center, bounds.width / 2
        lower, upper = ibp_forward_linear_jit(self.module.weight, self.module.bias, center, diff)

        return IntervalBounds(bounds.region, lower, upper)

    def propagate_size(self, in_size):
        return self.module.weight.size(-2)


class FixedLinear(nn.Linear):
    def __init__(self, weight, bias=None):
        super().__init__(weight.size(-1), weight.size(-2), bias=bias is not None)

        del self.weight
        self.register_buffer('weight', weight)

        if bias is not None:
            del self.bias
            self.register_buffer('bias', bias)


class ElementWiseLinear(nn.Module):
    def __init__(self, a, b=None):
        super().__init__()

        self.register_buffer('a', torch.as_tensor(a))

        if b is not None:
            b = torch.as_tensor(b)
        self.register_buffer('b', b)

    def forward(self, x):
        x = self.a * x
        if self.b is not None:
            x = x + self.b

        return x


def crown_backward_elementwise_linear_jit(a: torch.Tensor, b: Optional[Union[torch.Tensor, float]], W_tilde: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    if b is None:
        b = torch.tensor(0.0, device=W_tilde.device, dtype=W_tilde.dtype)
    elif b.dim() > 0:
        b = b.unsqueeze(-2).matmul(W_tilde.transpose(-1, -2)).squeeze(-2)
    else:
        b = b * W_tilde.sum(dim=-1)

    if a.dim() > 0:
        a = a.unsqueeze(-2)

    W_tilde = W_tilde * a

    return W_tilde, b


class BoundElementWiseLinear(BoundModule):

    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        if linear_bounds.lower is None:
            lower = None
        else:
            lower = crown_backward_elementwise_linear_jit(self.module.a, self.module.b, linear_bounds.lower[0])
            lower = (lower[0], lower[1] + linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = crown_backward_elementwise_linear_jit(self.module.a, self.module.b, linear_bounds.upper[0])
            upper = (upper[0], upper[1] + linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    @assert_bound_order
    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        lower, upper = self(bounds.lower), self(bounds.upper)
        lower, upper = torch.min(lower, upper), torch.max(lower, upper)

        return IntervalBounds(bounds.region, lower, upper)

    def propagate_size(self, in_size):
        return in_size



import abc
import logging
from functools import wraps
from typing import Tuple

import torch
from torch import nn
import numpy as np

from .general import BoundModule
from .bounds import LinearBounds, IntervalBounds
from .util import TensorFunction, clip_param_to_range_, proj_grad_to_range_

logger = logging.getLogger(__name__)


def assert_bound_order(func, position=0, keyword='preactivation'):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if len(args) > position:
            bounds = args[position]
        else:
            bounds = kwargs[keyword]

        assert torch.isnan(bounds.lower).any() or torch.isnan(bounds.upper).any() or \
               torch.all(bounds.lower <= bounds.upper + 1e-6)

        return func(self, *args, **kwargs)

    return wrapper


@torch.jit.script
def crown_backward_act_jit(W_tilde: torch.Tensor, alpha: Tuple[torch.Tensor, torch.Tensor], beta: Tuple[torch.Tensor, torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor]:
    _lambda = torch.where(W_tilde < 0, alpha[0].unsqueeze(-2), alpha[1].unsqueeze(-2))
    _delta = torch.where(W_tilde < 0, beta[0].unsqueeze(-2), beta[1].unsqueeze(-2))

    bias = torch.sum(W_tilde * _delta, dim=-1)
    W_tilde = W_tilde * _lambda

    return W_tilde, bias


class BoundActivation(BoundModule, abc.ABC):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.input_bounds = None

        self.alpha_lower, self.beta_lower = None, None
        self.alpha_upper, self.beta_upper = None, None

        self.bounded = False
        self.size = None

    @abc.abstractmethod
    def alpha_beta(self, preactivation):
        raise NotImplementedError()

    @property
    def need_relaxation(self):
        return not self.bounded

    def set_relaxation(self, linear_bounds):
        interval_bounds = linear_bounds.concretize()
        interval_bounds = IntervalBounds(
            linear_bounds.region,
            torch.max(interval_bounds.lower, self.input_bounds.lower),
            torch.min(interval_bounds.upper, self.input_bounds.upper)
        )

        self.alpha_beta(preactivation=interval_bounds)
        self.bounded = True

    def backward_relaxation(self, region):
        assert self.size is not None

        linear_bounds = self.initial_linear_bounds(region, self.size)
        return linear_bounds, self

    def clear_relaxation(self):
        self.input_bounds = None

        self.alpha_lower, self.beta_lower = None, None
        self.alpha_upper, self.beta_upper = None, None

        self.bounded = False

    def crown_backward(self, linear_bounds, optimize):
        assert self.bounded

        alpha_lower, alpha_upper = self.alpha_lower.detach().clone(), self.alpha_upper.detach().clone()
        beta_lower, beta_upper = self.beta_lower.detach().clone(), self.beta_upper.detach().clone()

        if optimize:
            alpha_lower, alpha_upper, beta_lower, beta_upper = \
                self.parameterize_alpha_beta(alpha_lower, alpha_upper, beta_lower, beta_upper)

        # NOTE: The order of alpha and beta are deliberately reverse - this is not a mistake!
        if linear_bounds.lower is None:
            lower = None
        else:
            alpha = alpha_upper, alpha_lower
            beta = beta_upper, beta_lower

            lower = crown_backward_act_jit(linear_bounds.lower[0], alpha, beta)
            lower = (lower[0], lower[1] + linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            alpha = alpha_lower, alpha_upper
            beta = beta_lower, beta_upper
            upper = crown_backward_act_jit(linear_bounds.upper[0], alpha, beta)
            upper = (upper[0], upper[1] + linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    @assert_bound_order
    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        if save_relaxation:
            self.alpha_beta(preactivation=bounds)
            self.bounded = True

        if save_input_bounds:
            self.input_bounds = bounds

        return IntervalBounds(bounds.region, self.module(bounds.lower), self.module(bounds.upper))

    def propagate_size(self, in_size):
        self.size = in_size
        return in_size

    @abc.abstractmethod
    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        raise NotImplementedError()


def regimes(lower: torch.Tensor, upper: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    zero_width = torch.isclose(lower, upper, rtol=0.0, atol=1e-8)
    n = (~zero_width) & (upper <= 0)
    p = (~zero_width) & (0 <= lower)
    np = (~zero_width) & (lower < 0) & (0 < upper)

    return zero_width, n, p, np


class BoundReLU(BoundActivation):
    def __init__(self, module, factory, adaptive_relu=True, **kwargs):
        super().__init__(module, factory, **kwargs)
        self.adaptive_relu = adaptive_relu

        self.unstable_lower, self.unstable_slope_lower = None, None

    def clear_relaxation(self):
        super().clear_relaxation()
        self.unstable_lower, self.unstable_slope_lower = None, None

    @assert_bound_order
    def alpha_beta(self, preactivation):
        """
        See the source below for information on what adaptive ReLU means and how regimes of the input
        are exploited to compute bounds.

        @misc{zhang2018efficient,
              title={Efficient Neural Network Robustness Certification with General Activation Functions},
              author={Huan Zhang and Tsui-Wei Weng and Pin-Yu Chen and Cho-Jui Hsieh and Luca Daniel},
              year={2018},
              eprint={1811.00866},
              archivePrefix={arXiv},
              primaryClass={cs.LG}
        }

        :param self:
        :param preactivation:
        """
        lower, upper = preactivation.lower, preactivation.upper
        zero_width, n, p, np = regimes(lower, upper)

        self.alpha_lower, self.beta_lower = torch.zeros_like(lower), torch.zeros_like(lower)
        self.alpha_upper, self.beta_upper = torch.zeros_like(lower), torch.zeros_like(lower)

        # Use upper and lower in the bias to account for a small numerical difference between lower and upper
        # which ought to be negligible, but may still be present due to torch.isclose.
        self.alpha_lower[zero_width], self.beta_lower[zero_width] = 0, self(lower[zero_width])
        self.alpha_upper[zero_width], self.beta_upper[zero_width] = 0, self(upper[zero_width])

        self.alpha_lower[n], self.beta_lower[n] = 0, 0
        self.alpha_upper[n], self.beta_upper[n] = 0, 0

        self.alpha_lower[p], self.beta_lower[p] = 1, 0
        self.alpha_upper[p], self.beta_upper[p] = 1, 0

        lower, upper = lower[np], upper[np]

        z = upper / (upper - lower)
        if self.adaptive_relu:
            # Utilize that bool->float conversion is true=1 and false=0
            a = (upper >= torch.abs(lower)).to(lower.dtype)
        else:
            a = z

        self.alpha_lower[np], self.beta_lower[np] = a, 0
        self.alpha_upper[np], self.beta_upper[np] = z, -lower * z

        # Allow parameterization
        # Save mask
        self.unstable_lower = np
        # Optimization variables [0, 1] - detach, clone, and require grad to perform back prop and optimization
        # Be sure that slope != 0 as that would result in zero gradient
        self.unstable_slope_lower = z.detach().clone().requires_grad_()

    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        if self.unstable_lower is not None:
            alpha_lower[self.unstable_lower] = self.unstable_slope_lower
        else:
            logger.warning('ReLU bound not parameterized but expected to')

        return alpha_lower, alpha_upper, beta_lower, beta_upper

    def bound_parameters(self):
        if self.unstable_lower is not None:
            yield self.unstable_slope_lower
        else:
            logger.warning('ReLU bound not parameterized but expected to')

    def clip_params(self):
        clip_param_to_range_(self.unstable_slope_lower, (0.0, 1.0))

    def project_grads(self):
        proj_grad_to_range_(self.unstable_slope_lower, (0.0, 1.0))


class BoundSigmoid(BoundActivation):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def clear_relaxation(self):
        super().clear_relaxation()
        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def derivative(self, x):
        return torch.sigmoid(x) * (1 - torch.sigmoid(x))

    @assert_bound_order
    def alpha_beta(self, preactivation):
        """
            Function to compute upper and lower bounds for S-shaped activation functions.
            Assumptions:
            - :LB < :UB
            - :derivative is the derivative of :func
            - :func is S-/sigmoid shaped

            See the source below for information on how regimes of the input are exploited to compute bounds.

            @misc{zhang2018efficient,
                  title={Efficient Neural Network Robustness Certification with General Activation Functions},
                  author={Huan Zhang and Tsui-Wei Weng and Pin-Yu Chen and Cho-Jui Hsieh and Luca Daniel},
                  year={2018},
                  eprint={1811.00866},
                  archivePrefix={arXiv},
                  primaryClass={cs.LG}
            }

        :param self:
        :param preactivation:
        """
        lower, upper = preactivation.lower, preactivation.upper
        zero_width, n, p, np = regimes(lower, upper)

        self.alpha_lower, self.beta_lower = torch.zeros_like(lower), torch.zeros_like(lower)
        self.alpha_upper, self.beta_upper = torch.zeros_like(lower), torch.zeros_like(lower)

        # Use upper and lower in the bias to account for a small numerical difference between lower and upper
        # which ought to be negligible, but may still be present due to torch.isclose.
        self.alpha_lower[zero_width], self.beta_lower[zero_width] = 0, self(lower[zero_width])
        self.alpha_upper[zero_width], self.beta_upper[zero_width] = 0, self(upper[zero_width])

        lower_act, upper_act = self(lower), self(upper)
        lower_prime, upper_prime = self.derivative(lower), self.derivative(upper)

        d = (lower + upper) * 0.5  # Let d be the midpoint of the two bounds
        d_act = self(d)
        d_prime = self.derivative(d)

        slope = (upper_act - lower_act) / (upper - lower)

        def add_linear(alpha, beta, mask, a, x, y, a_mask=True):
            if a_mask:
                a = a[mask]

            alpha[mask] = a
            beta[mask] = y[mask] - a * x[mask]

        ###################
        # Negative regime #
        ###################
        # Upper bound
        # - Exact slope between lower and upper
        add_linear(self.alpha_upper, self.beta_upper, mask=n, a=slope, x=upper, y=upper_act)

        # Lower bound
        # - d = (lower + upper) / 2 for midpoint
        # - Slope is sigma'(d) and it has to cross through sigma(d)
        add_linear(self.alpha_lower, self.beta_lower, mask=n, a=d_prime, x=d, y=d_act)

        # Allow parameterization
        # Save mask
        self.unstable_lower = n
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        self.unstable_d_lower = d[n].detach().clone().requires_grad_()
        # Save ranges to clip (aka. PGD)
        self.unstable_range_lower = lower[n], upper[n]

        ###################
        # Positive regime #
        ###################
        # Lower bound
        # - Exact slope between lower and upper
        add_linear(self.alpha_lower, self.beta_lower, mask=p, a=slope, x=lower, y=lower_act)

        # Upper bound
        # - d = (lower + upper) / 2 for midpoint
        # - Slope is sigma'(d) and it has to cross through sigma(d)
        add_linear(self.alpha_upper, self.beta_upper, mask=p, a=d_prime, x=d, y=d_act)

        # Allow parameterization
        # Save mask
        self.unstable_upper = p
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        self.unstable_d_upper = d[p].detach().clone().requires_grad_()
        # Save ranges to clip (aka. PGD)
        self.unstable_range_upper = lower[p], upper[p]

        #################
        # Crossing zero #
        #################
        # Upper bound #
        # If tangent to upper is below lower, then take direct slope between lower and upper
        direct = np & (slope <= upper_prime)
        add_linear(self.alpha_upper, self.beta_upper, mask=direct, a=slope, x=lower, y=lower_act)

        # Else use bisection to find upper bound on slope.
        implicit = np & (slope > upper_prime)

        if torch.any(implicit):
            implicit_lower, implicit_upper = lower[implicit], upper[implicit]
            implicit_lower_act = self(implicit_lower)

            def f_upper(d: torch.Tensor) -> torch.Tensor:
                a_slope = (self(d) - implicit_lower_act) / (d - implicit_lower)
                a_derivative = self.derivative(d)
                return a_slope - a_derivative

            # Bisection will return left and right bounds for d s.t. f_upper(d) is zero
            # Derivative of left bound will over-approximate the slope - hence a true bound
            d_upper, _ = bisection(torch.zeros_like(implicit_upper), implicit_upper, f_upper)
            # Slope has to attach to (lower, sigma(lower))
            add_linear(self.alpha_upper, self.beta_upper, mask=implicit, a=self.derivative(d_upper), x=lower, y=lower_act, a_mask=False)

        # Lower bound #
        # If tangent to lower is above upper, then take direct slope between lower and upper
        direct = np & (slope <= lower_prime)
        add_linear(self.alpha_lower, self.beta_lower, mask=direct, a=slope, x=upper, y=upper_act)

        # Else use bisection to find upper bound on slope.
        implicit = np & (slope > lower_prime)

        if torch.any(implicit):
            implicit_lower, implicit_upper = lower[implicit], upper[implicit]
            implicit_upper_act = self(implicit_upper)

            def f_lower(d: torch.Tensor) -> torch.Tensor:
                a_slope = (implicit_upper_act - self(d)) / (implicit_upper - d)
                a_derivative = self.derivative(d)
                return a_derivative - a_slope

            # Bisection will return left and right bounds for d s.t. f_lower(d) is zero
            # Derivative of right bound will over-approximate the slope - hence a true bound
            _, d_lower = bisection(implicit_lower, torch.zeros_like(implicit_lower), f_lower)
            # Slope has to attach to (upper, sigma(upper))
            add_linear(self.alpha_lower, self.beta_lower, mask=implicit, a=self.derivative(d_lower), x=upper, y=upper_act, a_mask=False)

    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Sigmoid/tanh bound not parameterized but expected to')

        # Use implicit parameterization (i.e. store d [point where touching the curve], and not alpha)
        def add_linear(alpha, beta, mask, x):
            a = self.derivative(x)
            y = self(x)

            alpha[mask] = a
            beta[mask] = y - a * x

        add_linear(alpha_lower, beta_lower, mask=self.unstable_lower, x=self.unstable_d_lower)
        add_linear(alpha_upper, beta_upper, mask=self.unstable_upper, x=self.unstable_d_upper)

        return alpha_lower, alpha_upper, beta_lower, beta_upper

    def bound_parameters(self):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Sigmoid/tanh bound not parameterized but expected to')

        yield self.unstable_d_lower
        yield self.unstable_d_upper

    def clip_params(self):
        clip_param_to_range_(self.unstable_d_lower, self.unstable_range_lower)
        clip_param_to_range_(self.unstable_d_upper, self.unstable_range_upper)

    def project_grads(self):
        proj_grad_to_range_(self.unstable_d_lower, self.unstable_range_lower)
        proj_grad_to_range_(self.unstable_d_upper, self.unstable_range_upper)


def bisection(l: torch.Tensor, u: torch.Tensor, f: TensorFunction, num_iter: int = 10) -> Tuple[torch.Tensor, torch.Tensor]:
    l, u = l.detach().clone(), u.detach().clone()
    midpoint = (l + u) / 2

    for _ in range(num_iter):
        y = f(midpoint)

        msk = y <= 0
        l[msk] = midpoint[msk]
        u[~msk] = midpoint[~msk]

        midpoint = (l + u) / 2

    return l, u


class BoundTanh(BoundSigmoid):
    # Tanh is actually just a scaled sigmoid, so let's reuse that code
    def derivative(self, x):
        return 1 - torch.tanh(x) ** 2


class BoundIdentity(BoundModule):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        return linear_bounds

    @assert_bound_order
    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        return bounds

    def propagate_size(self, in_size):
        return in_size


class Exp(nn.Module):
    def forward(self, x):
        return x.exp()


class BoundExp(BoundActivation):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None

    def clear_relaxation(self):
        super().clear_relaxation()
        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None

    def derivative(self, x):
        return x.exp()

    @assert_bound_order
    def alpha_beta(self, preactivation):
        lower, upper = preactivation.lower, preactivation.upper
        zero_width, n, p, np = regimes(lower, upper)
        all = (n | p | np)

        self.alpha_lower, self.beta_lower = torch.zeros_like(lower), torch.zeros_like(lower)
        self.alpha_upper, self.beta_upper = torch.zeros_like(lower), torch.zeros_like(lower)

        # Use upper and lower in the bias to account for a small numerical difference between lower and upper
        # which ought to be negligible, but may still be present due to torch.isclose.
        self.alpha_lower[zero_width], self.beta_lower[zero_width] = 0, self(lower[zero_width])
        self.alpha_upper[zero_width], self.beta_upper[zero_width] = 0, self(upper[zero_width])

        lower_act, upper_act = self(lower), self(upper)

        d = (lower + upper) * 0.5  # Let d be the midpoint of the two bounds
        d_act = self(d)
        d_prime = self.derivative(d)

        slope = (upper_act - lower_act) / (upper - lower)

        def add_linear(alpha, beta, a, x, y, a_mask=True):
            if a_mask:
                a = a[all]

            alpha[all] = a
            beta[all] = y[all] - a * x[all]

        # Upper bound
        # - Exact slope between lower and upper
        add_linear(self.alpha_upper, self.beta_upper, a=slope, x=lower, y=lower_act)

        # Lower bound
        # - d = (lower + upper) / 2 for midpoint
        # - Slope is sigma'(d) and it has to cross through sigma(d)
        add_linear(self.alpha_lower, self.beta_lower, a=d_prime, x=d, y=d_act)

        # Allow parameterization
        # Save mask
        self.unstable_lower = all
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        self.unstable_d_lower = d[all].detach().clone().requires_grad_()
        # Save ranges to clip (aka. PGD)
        self.unstable_range_lower = lower[all], upper[all]

    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        if self.unstable_lower is None:
            logger.warning('Exp not parameterized but expected to')

        # Use implicit parameterization (i.e. store d [point where touching the curve], and not alpha)
        a = self.derivative(self.unstable_d_lower)
        y = self(self.unstable_d_lower)

        alpha_lower[self.unstable_lower] = a
        beta_lower[self.unstable_lower] = y - a * self.unstable_d_lower

        return alpha_lower, alpha_upper, beta_lower, beta_upper

    def bound_parameters(self):
        if self.unstable_lower is None:
            logger.warning('Exp bound not parameterized but expected to')

        yield self.unstable_d_lower

    def clip_params(self):
        clip_param_to_range_(self.unstable_d_lower, self.unstable_range_lower)

    def project_grads(self):
        proj_grad_to_range_(self.unstable_d_lower, self.unstable_range_lower)


class Log(nn.Module):
    def forward(self, x):
        # Assumption: x > 0 (as log is only defined for x > 0)
        return x.log()


class BoundLog(BoundActivation):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def clear_relaxation(self):
        super().clear_relaxation()
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def derivative(self, x):
        # As we already assume x > 0, this is well-defined
        return 1 / x

    @assert_bound_order
    def alpha_beta(self, preactivation):
        lower, upper = preactivation.lower, preactivation.upper
        assert torch.all(lower > 0)

        zero_width, n, p, np = regimes(lower, upper)
        all = (n | p | np)

        self.alpha_lower, self.beta_lower = torch.zeros_like(lower), torch.zeros_like(lower)
        self.alpha_upper, self.beta_upper = torch.zeros_like(lower), torch.zeros_like(lower)

        # Use upper and lower in the bias to account for a small numerical difference between lower and upper
        # which ought to be negligible, but may still be present due to torch.isclose.
        self.alpha_lower[zero_width], self.beta_lower[zero_width] = 0, self(lower[zero_width])
        self.alpha_upper[zero_width], self.beta_upper[zero_width] = 0, self(upper[zero_width])

        lower_act, upper_act = self(lower), self(upper)

        d = (lower + upper) * 0.5  # Let d be the midpoint of the two bounds
        d_act = self(d)
        d_prime = self.derivative(d)

        slope = (upper_act - lower_act) / (upper - lower)

        def add_linear(alpha, beta, a, x, y, a_mask=True):
            if a_mask:
                a = a[all]

            alpha[all] = a
            beta[all] = y[all] - a * x[all]

        # Lower bound
        # - Exact slope between lower and upper
        add_linear(self.alpha_lower, self.beta_lower, a=slope, x=lower, y=lower_act)

        # Upper bound
        # - d = (lower + upper) / 2 for midpoint
        # - Slope is sigma'(d) and it has to cross through sigma(d)
        add_linear(self.alpha_upper, self.beta_upper, a=d_prime, x=d, y=d_act)

        # Allow parameterization
        # Save mask
        self.unstable_upper = all
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        self.unstable_d_upper = d[all].detach().clone().requires_grad_()
        # Save ranges to clip (aka. PGD)
        self.unstable_range_upper = lower[all], upper[all]

    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        if self.unstable_upper is None:
            logger.warning('Log not parameterized but expected to')

        # Use implicit parameterization (i.e. store d [point where touching the curve], and not alpha)
        a = self.derivative(self.unstable_d_upper)
        y = self(self.unstable_d_upper)

        alpha_upper[self.unstable_upper] = a
        beta_upper[self.unstable_upper] = y - a * self.unstable_d_upper

        return alpha_lower, alpha_upper, beta_lower, beta_upper

    def bound_parameters(self):
        if self.unstable_upper is None:
            logger.warning('Log bound not parameterized but expected to')

        yield self.unstable_d_upper

    def clip_params(self):
        clip_param_to_range_(self.unstable_d_upper, self.unstable_range_upper)

    def project_grads(self):
        proj_grad_to_range_(self.unstable_d_upper, self.unstable_range_upper)


class Reciprocal(nn.Module):
    def forward(self, x):
        return 1 / x


class BoundReciprocal(BoundActivation):
    # Only negative or positive. If crossing zero then no linear relaxation can exist as it goes to +- \infty

    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def clear_relaxation(self):
        super().clear_relaxation()
        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def derivative(self, x):
        return -1 / (x ** 2)

    @assert_bound_order
    def alpha_beta(self, preactivation):
        lower, upper = preactivation.lower, preactivation.upper
        zero_width, n, p, np = regimes(lower, upper)
        assert not torch.any(np), 'Input to reciprocal cannot cross zero as no linear bound can exist'

        self.alpha_lower, self.beta_lower = torch.zeros_like(lower), torch.zeros_like(lower)
        self.alpha_upper, self.beta_upper = torch.zeros_like(lower), torch.zeros_like(lower)

        # Use upper and lower in the bias to account for a small numerical difference between lower and upper
        # which ought to be negligible, but may still be present due to torch.isclose.
        self.alpha_lower[zero_width], self.beta_lower[zero_width] = 0, self(lower[zero_width])
        self.alpha_upper[zero_width], self.beta_upper[zero_width] = 0, self(upper[zero_width])

        lower_act, upper_act = self(lower), self(upper)

        d = (lower + upper) * 0.5  # Let d be the midpoint of the two bounds
        d_act = self(d)
        d_prime = self.derivative(d)

        slope = (upper_act - lower_act) / (upper - lower)

        def add_linear(alpha, beta, mask, a, x, y, a_mask=True):
            if a_mask:
                a = a[mask]

            alpha[mask] = a
            beta[mask] = y[mask] - a * x[mask]

        ###################
        # Negative regime #
        ###################
        # Lower bound
        # - Exact slope between lower and upper
        add_linear(self.alpha_lower, self.beta_lower, mask=n, a=slope, x=upper, y=upper_act)

        # Upper bound
        # - d = (lower + upper) / 2 for midpoint
        # - Slope is sigma'(d) and it has to cross through sigma(d)
        add_linear(self.alpha_upper, self.alpha_upper, mask=n, a=d_prime, x=d, y=d_act)

        # Allow parameterization
        # Save mask
        self.unstable_upper = n
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        self.unstable_d_upper = d[n].detach().clone().requires_grad_()
        # Save ranges to clip (aka. PGD)
        self.unstable_range_upper = lower[n], upper[n]

        ###################
        # Positive regime #
        ###################
        # Upper bound
        # - Exact slope between lower and upper
        add_linear(self.alpha_upper, self.beta_upper, mask=p, a=slope, x=lower, y=lower_act)

        # Lower bound
        # - d = (lower + upper) / 2 for midpoint
        # - Slope is sigma'(d) and it has to cross through sigma(d)
        add_linear(self.alpha_lower, self.beta_lower, mask=p, a=d_prime, x=d, y=d_act)

        # Allow parameterization
        # Save mask
        self.unstable_lower = p
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        self.unstable_d_lower = d[p].detach().clone().requires_grad_()
        # Save ranges to clip (aka. PGD)
        self.unstable_range_lower = lower[p], upper[p]

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        if save_relaxation:
            self.alpha_beta(preactivation=bounds)
            self.bounded = True

        if save_input_bounds:
            self.input_bounds = bounds

        lower_act = self.module(bounds.lower)
        upper_act = self.module(bounds.upper)

        lower = torch.min(lower_act, upper_act)
        upper = torch.max(lower_act, upper_act)

        return IntervalBounds(bounds.region, lower, upper)

    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Reciprocal bound not parameterized but expected to')

        # Use implicit parameterization (i.e. store d [point where touching the curve], and not alpha)
        def add_linear(alpha, beta, mask, x):
            a = self.derivative(x)
            y = self(x)

            alpha[mask] = a
            beta[mask] = y - a * x

        add_linear(alpha_lower, beta_lower, mask=self.unstable_lower, x=self.unstable_d_lower)
        add_linear(alpha_upper, beta_upper, mask=self.unstable_upper, x=self.unstable_d_upper)

        return alpha_lower, alpha_upper, beta_lower, beta_upper

    def bound_parameters(self):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Reciprocal bound not parameterized but expected to')

        yield self.unstable_d_lower
        yield self.unstable_d_upper

    def clip_params(self):
        clip_param_to_range_(self.unstable_d_lower, self.unstable_range_lower)
        clip_param_to_range_(self.unstable_d_upper, self.unstable_range_upper)

    def project_grads(self):
        proj_grad_to_range_(self.unstable_d_lower, self.unstable_range_lower)
        proj_grad_to_range_(self.unstable_d_upper, self.unstable_range_upper)


class Sin(nn.Module):
    def forward(self, x):
        return x.sin()


def sine_like_regimes(lower, upper, period, zero_increasing):
    zero_width = torch.isclose(lower, upper, rtol=0.0, atol=1e-8)

    half_period = upper - lower >= period / 2

    zero_shifted_lower = lower - zero_increasing
    zero_shifted_upper = upper - zero_increasing

    shift = torch.div(zero_shifted_lower, period, rounding_mode='floor') * period
    shifted_lower = zero_shifted_lower - shift
    shifted_upper = zero_shifted_upper - shift

    increasing_region1 = (shifted_upper <= (1 / 4) * period)
    increasing_region2 = (shifted_lower >= (3 / 4) * period) & (shifted_upper <= (5 / 4) * period)
    increasing_region3 = (shifted_lower >= (7 / 4) * period)
    increasing = (~zero_width) & (~half_period) & (increasing_region1 | increasing_region2 | increasing_region3)

    increasing_region1 = (shifted_lower >= (3 / 4) * period) & (shifted_upper <= period)
    increasing_region2 = (shifted_lower >= (7 / 4) * period)
    increasing_lower_curve = increasing & (increasing_region1 | increasing_region2)

    increasing_region1 = (shifted_upper <= (1 / 4) * period)
    increasing_region2 = (shifted_lower >= period) & (shifted_upper <= (5 / 4) * period)
    increasing_upper_curve = increasing & (increasing_region1 | increasing_region2)

    increasing_full_region = increasing & (~increasing_lower_curve) & (~increasing_upper_curve)
    increasing = increasing, (increasing_lower_curve, increasing_upper_curve, increasing_full_region)

    decreasing_region1 = (shifted_lower >= (1 / 4) * period) & (shifted_upper <= (3 / 4) * period)
    decreasing_region2 = (shifted_lower >= (5 / 4) * period) & (shifted_upper <= (7 / 4) * period)
    decreasing = (~zero_width) & (~half_period) & (decreasing_region1 | decreasing_region2)

    decreasing_region1 = (shifted_lower >= (2 / 4) * period) & (shifted_upper <= (3 / 4) * period)
    decreasing_region2 = (shifted_lower >= (6 / 4) * period) & (shifted_upper <= (7 / 4) * period)
    decreasing_lower_curve = decreasing & (decreasing_region1 | decreasing_region2)

    decreasing_region1 = (shifted_lower >= (1 / 4) * period) & (shifted_upper <= (2 / 4) * period)
    decreasing_region2 = (shifted_lower >= (5 / 4) * period) & (shifted_upper <= (6 / 4) * period)
    decreasing_upper_curve = decreasing & (decreasing_region1 | decreasing_region2)

    decreasing_full_region = decreasing & (~decreasing_lower_curve) & (~decreasing_upper_curve)
    decreasing = decreasing, (decreasing_lower_curve, decreasing_upper_curve, decreasing_full_region)

    crossing_peak_region1 = (shifted_lower < (1 / 4) * period) & (shifted_upper > (1 / 4) * period)
    crossing_peak_region2 = (shifted_lower < (5 / 4) * period) & (shifted_upper > (5 / 4) * period)
    crossing_peak = (~zero_width) & (~half_period) & (crossing_peak_region1 | crossing_peak_region2)

    crossing_trough_region1 = (shifted_lower < (3 / 4) * period) & (shifted_upper > (3 / 4) * period)
    crossing_trough_region2 = (shifted_lower < (7 / 4) * period) & (shifted_upper > (7 / 4) * period)
    crossing_trough = (~zero_width) & (~half_period) & (crossing_trough_region1 | crossing_trough_region2)

    return zero_width, half_period, increasing, decreasing, crossing_peak, crossing_trough


class BoundSin(BoundActivation):
    period = 2 * np.pi
    zero_increasing = 0

    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def clear_relaxation(self):
        super().clear_relaxation()
        self.unstable_lower, self.unstable_d_lower, self.unstable_range_lower = None, None, None
        self.unstable_upper, self.unstable_d_upper, self.unstable_range_upper = None, None, None

    def derivative(self, x):
        return x.cos()

    @assert_bound_order
    def alpha_beta(self, preactivation):
        lower, upper = preactivation.lower, preactivation.upper

        zero_width, half_period, (_, increasing), (_, decreasing), crossing_peak, crossing_trough = \
            sine_like_regimes(lower, upper, period=self.period, zero_increasing=self.zero_increasing)
        increasing_lower_curve, increasing_upper_curve, increasing_full_region = increasing
        decreasing_lower_curve, decreasing_upper_curve, decreasing_full_region = decreasing

        self.alpha_lower, self.beta_lower = torch.zeros_like(lower), torch.zeros_like(lower)
        self.alpha_upper, self.beta_upper = torch.zeros_like(lower), torch.zeros_like(lower)

        # Use upper and lower in the bias to account for a small numerical difference between lower and upper
        # which ought to be negligible, but may still be present due to torch.isclose.
        self.alpha_lower[zero_width], self.beta_lower[zero_width] = 0, self(lower[zero_width])
        self.alpha_upper[zero_width], self.beta_upper[zero_width] = 0, self(upper[zero_width])

        lower_act, upper_act = self(lower), self(upper)
        lower_prime, upper_prime = self.derivative(lower), self.derivative(upper)

        d = (lower + upper) * 0.5  # Let d be the midpoint of the two bounds
        d_act = self(d)
        d_prime = self.derivative(d)

        slope = (upper_act - lower_act) / (upper - lower)

        ones = torch.ones_like(lower)
        zeros = torch.zeros_like(lower)

        def add_linear(alpha, beta, mask, a, x, y, a_mask=True):
            if a_mask:
                a = a[mask]

            alpha[mask] = a
            beta[mask] = y[mask] - a * x[mask]

        ##################
        # >= Half period #
        ##################
        # Lower bound
        # - Flat line = -1
        add_linear(self.alpha_lower, self.beta_lower, mask=half_period, a=zeros, x=zeros, y=-ones)

        # Upper bound
        # - Flat line = +1
        add_linear(self.alpha_upper, self.beta_upper, mask=half_period, a=zeros, x=zeros, y=ones)

        ###############
        # Lower curve #
        ###############
        lower_curve = increasing_lower_curve | decreasing_lower_curve | crossing_trough

        # Upper bound
        # - Exact slope between lower and upper
        add_linear(self.alpha_upper, self.beta_upper, mask=lower_curve, a=slope, x=lower, y=lower_act)

        # Lower bound
        # - d = (lower + upper) / 2 for midpoint
        # - Slope is sigma'(d) and it has to cross through sigma(d)
        add_linear(self.alpha_lower, self.beta_lower, mask=lower_curve, a=d_prime, x=d, y=d_act)

        # Allow parameterization
        # Save mask
        self.unstable_lower = lower_curve
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        self.unstable_d_lower = d[lower_curve].detach().clone().requires_grad_()
        # Save ranges to clip (aka. PGD)
        self.unstable_range_lower = lower[lower_curve], upper[lower_curve]

        ###############
        # Upper curve #
        ###############
        upper_curve = increasing_upper_curve | decreasing_upper_curve | crossing_peak

        # Lower bound
        # - Exact slope between lower and upper
        add_linear(self.alpha_lower, self.beta_lower, mask=upper_curve, a=slope, x=upper, y=upper_act)

        # Upper bound
        # - d = (lower + upper) / 2 for midpoint
        # - Slope is sigma'(d) and it has to cross through sigma(d)
        add_linear(self.alpha_upper, self.beta_upper, mask=upper_curve, a=d_prime, x=d, y=d_act)

        # Allow parameterization
        # Save mask
        self.unstable_upper = upper_curve
        # Optimization variables - detach, clone, and require grad to perform back prop and optimization
        self.unstable_d_upper = d[upper_curve].detach().clone().requires_grad_()
        # Save ranges to clip (aka. PGD)
        self.unstable_range_upper = lower[upper_curve], upper[upper_curve]

        # ##########################
        # Increasing full region #
        ##########################
        # Upper bound #
        # If tangent to upper is below lower, then take direct slope between lower and upper
        direct = increasing_full_region & (slope <= upper_prime)
        add_linear(self.alpha_upper, self.beta_upper, mask=direct, a=slope, x=lower, y=lower_act)

        # Else use bisection to find upper bound on slope.
        implicit = increasing_full_region & (slope > upper_prime)

        if torch.any(implicit):
            implicit_lower, implicit_upper = lower[implicit], upper[implicit]
            implicit_lower_act = self(implicit_lower)

            def f_upper(d: torch.Tensor) -> torch.Tensor:
                a_slope = (self(d) - implicit_lower_act) / (d - implicit_lower)
                a_derivative = self.derivative(d)
                return a_slope - a_derivative

            # Bisection will return left and right bounds for d s.t. f_upper(d) is zero
            # Derivative of left bound will over-approximate the slope - hence a true bound
            bisection_lower = implicit_lower + self.period / 4 - torch.remainder(implicit_lower - self.zero_increasing, self.period / 4)
            d_upper, _ = bisection(bisection_lower, implicit_upper, f_upper)
            # Slope has to attach to (lower, sigma(lower))
            add_linear(self.alpha_upper, self.beta_upper, mask=implicit, a=self.derivative(d_upper), x=lower, y=lower_act, a_mask=False)

        # Lower bound #
        # If tangent to lower is above upper, then take direct slope between lower and upper
        direct = increasing_full_region & (slope <= lower_prime)
        add_linear(self.alpha_lower, self.beta_lower, mask=direct, a=slope, x=upper, y=upper_act)

        # Else use bisection to find upper bound on slope.
        implicit = increasing_full_region & (slope > lower_prime)

        if torch.any(implicit):
            implicit_lower, implicit_upper = lower[implicit], upper[implicit]
            implicit_upper_act = self(implicit_upper)

            def f_lower(d: torch.Tensor) -> torch.Tensor:
                a_slope = (implicit_upper_act - self(d)) / (implicit_upper - d)
                a_derivative = self.derivative(d)
                return a_derivative - a_slope

            # Bisection will return left and right bounds for d s.t. f_lower(d) is zero
            # Derivative of right bound will over-approximate the slope - hence a true bound
            bisection_upper = implicit_upper - torch.remainder(implicit_upper - self.zero_increasing, self.period / 4)
            _, d_lower = bisection(implicit_lower, bisection_upper, f_lower)
            # Slope has to attach to (upper, sigma(upper))
            add_linear(self.alpha_lower, self.beta_lower, mask=implicit, a=self.derivative(d_lower), x=upper, y=upper_act, a_mask=False)

        ##########################
        # Decreasing full region #
        ##########################
        # Upper bound #
        # If tangent to lower is below upper, then take direct slope between lower and upper
        direct = decreasing_full_region & (slope >= lower_prime)
        add_linear(self.alpha_upper, self.beta_upper, mask=direct, a=slope, x=lower, y=lower_act)

        # Else use bisection to find upper bound on slope.
        implicit = decreasing_full_region & (slope < lower_prime)

        if torch.any(implicit):
            implicit_lower, implicit_upper = lower[implicit], upper[implicit]
            implicit_lower_act = self(implicit_lower)

            def f_upper(d: torch.Tensor) -> torch.Tensor:
                a_slope = (implicit_lower_act - self(d)) / (implicit_lower - d)
                a_derivative = self.derivative(d)
                return a_derivative - a_slope

            # Bisection will return left and right bounds for d s.t. f_upper(d) is zero
            # Derivative of right bound will over-approximate the slope - hence a true bound
            bisection_upper = implicit_upper - torch.remainder(implicit_upper - self.zero_increasing, self.period / 4)
            _, d_upper = bisection(implicit_lower, bisection_upper, f_upper)
            # Slope has to attach to (lower, sigma(lower))
            add_linear(self.alpha_upper, self.beta_upper, mask=implicit, a=self.derivative(d_upper), x=upper, y=upper_act, a_mask=False)

        # Lower bound #
        # If tangent to upper is above lower, then take direct slope between lower and upper
        direct = decreasing_full_region & (slope >= upper_prime)
        add_linear(self.alpha_lower, self.beta_lower, mask=direct, a=slope, x=lower, y=lower_act)

        # Else use bisection to find upper bound on slope.
        implicit = decreasing_full_region & (slope < upper_prime)

        if torch.any(implicit):
            implicit_lower, implicit_upper = lower[implicit], upper[implicit]
            implicit_upper_act = self(implicit_upper)

            def f_lower(d: torch.Tensor) -> torch.Tensor:
                a_slope = (self(d) - implicit_upper_act) / (d - implicit_upper)
                a_derivative = self.derivative(d)
                return a_slope - a_derivative

            # Bisection will return left and right bounds for d s.t. f_lower(d) is zero
            # Derivative of left bound will over-approximate the slope - hence a true bound
            bisection_lower = implicit_lower + self.period / 4 - torch.remainder(implicit_lower - self.zero_increasing, self.period / 4)
            d_lower, _ = bisection(bisection_lower, implicit_upper, f_lower)
            # Slope has to attach to (upper, sigma(upper))
            add_linear(self.alpha_lower, self.beta_lower, mask=implicit, a=self.derivative(d_lower), x=lower, y=lower_act, a_mask=False)

    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        if save_relaxation:
            self.alpha_beta(preactivation=bounds)
            self.bounded = True

        if save_input_bounds:
            self.input_bounds = bounds

        zero_width, half_period, (increasing, _), (decreasing, _), crossing_peak, crossing_trough = \
            sine_like_regimes(bounds.lower, bounds.upper, period=self.period, zero_increasing=self.zero_increasing)

        lower = torch.zeros_like(bounds.lower)
        upper = torch.zeros_like(bounds.upper)

        lower_act = self.module(bounds.lower)
        upper_act = self.module(bounds.upper)

        lower[zero_width] = torch.min(lower_act[zero_width], upper_act[zero_width])
        upper[zero_width] = torch.max(lower_act[zero_width], upper_act[zero_width])

        lower[half_period] = -1
        upper[half_period] = 1

        lower[increasing] = lower_act[increasing]
        upper[increasing] = upper_act[increasing]

        lower[decreasing] = upper_act[decreasing]
        upper[decreasing] = lower_act[decreasing]

        lower[crossing_peak] = torch.min(lower_act[crossing_peak], upper_act[crossing_peak])
        upper[crossing_peak] = 1

        lower[crossing_trough] = -1
        upper[crossing_trough] = torch.max(lower_act[crossing_trough], upper_act[crossing_trough])

        return IntervalBounds(bounds.region, lower, upper)

    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Sin/cos bound not parameterized but expected to')

        # Use implicit parameterization (i.e. store d [point where touching the curve], and not alpha)
        def add_linear(alpha, beta, mask, x):
            a = self.derivative(x)
            y = self(x)

            alpha[mask] = a
            beta[mask] = y - a * x

        add_linear(alpha_lower, beta_lower, mask=self.unstable_lower, x=self.unstable_d_lower)
        add_linear(alpha_upper, beta_upper, mask=self.unstable_upper, x=self.unstable_d_upper)

        return alpha_lower, alpha_upper, beta_lower, beta_upper

    def bound_parameters(self):
        if self.unstable_lower is None or self.unstable_upper is None:
            logger.warning('Sin/cos bound not parameterized but expected to')

        yield self.unstable_d_lower
        yield self.unstable_d_upper

    def clip_params(self):
        clip_param_to_range_(self.unstable_d_lower, self.unstable_range_lower)
        clip_param_to_range_(self.unstable_d_upper, self.unstable_range_upper)

    def project_grads(self):
        proj_grad_to_range_(self.unstable_d_lower, self.unstable_range_lower)
        proj_grad_to_range_(self.unstable_d_upper, self.unstable_range_upper)


class Cos(nn.Module):
    def forward(self, x):
        return x.cos()


class BoundCos(BoundSin):
    period = 2 * np.pi
    zero_increasing = -np.pi / 2

    def derivative(self, x):
        return -x.sin()



import torch


class HyperRectangle:
    def __init__(self, lower, upper):
        self.lower, self.upper = lower, upper

    @property
    def width(self):
        return self.upper - self.lower

    @property
    def center(self):
        return (self.upper + self.lower) / 2

    def __len__(self):
        return self.lower.size(0)

    @staticmethod
    def from_eps(x, eps):
        lower, upper = x - eps, x + eps
        return HyperRectangle(lower, upper)

    def __getitem__(self, item):
        return HyperRectangle(
            self.lower[item] if self.lower is not None else None,
            self.upper[item] if self.upper is not None else None
        )

    def to(self, *args, **kwargs):
        return HyperRectangle(
            self.lower.to(*args, **kwargs) if self.lower is not None else None,
            self.upper.to(*args, **kwargs) if self.upper is not None else None
        )

    def cpu(self):
        return self.to(torch.device('cpu'))


class IntervalBounds(HyperRectangle):
    def __init__(self, region, lower, upper):
        super().__init__(lower, upper)
        self.region = region

    def __getitem__(self, item):
        return IntervalBounds(
            self.region[item],
            self.lower[item] if self.lower is not None else None,
            self.upper[item] if self.upper is not None else None
        )

    def to(self, *args, **kwargs):
        return IntervalBounds(
            self.region.to(*args, **kwargs),
            self.lower.to(*args, **kwargs) if self.lower is not None else None,
            self.upper.to(*args, **kwargs) if self.upper is not None else None
        )

    def concretize(self):
        return self


class LinearBounds:
    def __init__(self, region, lower, upper):
        self.region = region
        self.lower, self.upper = lower, upper

    def concretize(self):
        center, diff = self.region.center, self.region.width / 2
        center, diff = center.unsqueeze(-2), diff.unsqueeze(-2)

        if self.lower is not None:
            slope, intercept = self.lower
            slope = slope.transpose(-1, -2)
            lower = center.matmul(slope) - diff.matmul(slope.abs())
            lower = lower.squeeze(-2) + intercept
        else:
            lower = None

        if self.upper is not None:
            slope, intercept = self.upper
            slope = slope.transpose(-1, -2)
            upper = center.matmul(slope) + diff.matmul(slope.abs())
            upper = upper.squeeze(-2) + intercept
        else:
            upper = None

        return IntervalBounds(self.region, lower, upper)

    def __len__(self):
        return len(self.region)

    def __getitem__(self, idx):
        if isinstance(idx, tuple) and len(idx) >= 2 and idx[0] == Ellipsis:
            bias_idx = idx[:-1]
            region_idx = idx[:-2] + idx[-1:]
        else:
            bias_idx = idx
            region_idx = idx

        return LinearBounds(
            self.region[region_idx],
            (self.lower[0][idx], self.lower[1][bias_idx]) if self.lower is not None else None,
            (self.upper[0][idx], self.upper[1][bias_idx]) if self.upper is not None else None
        )

    def to(self, *args, **kwargs):
        return LinearBounds(
            self.region.to(*args, **kwargs),
            (self.lower[0].to(*args, **kwargs), self.lower[1].to(*args, **kwargs)) if self.lower is not None else None,
            (self.upper[0].to(*args, **kwargs), self.upper[1].to(*args, **kwargs)) if self.upper is not None else None
        )

    def cpu(self):
        return self.to(torch.device('cpu'))



import logging
from typing import Tuple, Optional, Union

import torch
from torch import nn, Tensor

from bound_propagation import BoundActivation
from bound_propagation.activation import assert_bound_order
from bound_propagation.util import clip_param_to_range_, proj_grad_to_range_

logger = logging.getLogger(__name__)


class Clamp(nn.Module):
    def __init__(self, min=None, max=None):
        super().__init__()

        assert min is not None or max is not None

        self.min = min
        self.max = max

    def forward(self, x):
        return x.clamp(min=self.min, max=self.max)


def regimes(lower: torch.Tensor, upper: torch.Tensor, min: Optional[Union[Tensor, float]],
            max: Optional[Union[Tensor, float]]) -> Tuple[
    torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    flat_lower = torch.full_like(lower, False, dtype=torch.bool)
    flat_upper = torch.full_like(lower, False, dtype=torch.bool)
    slope = torch.full_like(lower, False, dtype=torch.bool)
    lower_bend = torch.full_like(lower, False, dtype=torch.bool)
    upper_bend = torch.full_like(lower, False, dtype=torch.bool)
    full_range = torch.full_like(lower, False, dtype=torch.bool)
    zero_width = torch.isclose(lower, upper, rtol=0.0, atol=1e-8)

    tautology = torch.full_like(lower, True, dtype=torch.bool)

    if min is not None:
        flat_lower |= (~zero_width) & (upper <= min)
        lower_bend |= (~zero_width) & (lower < min) & (upper > min) & (tautology if max is None else (upper < max))
    else:
        slope |= (~zero_width) & (upper <= max)

    if max is not None:
        flat_upper |= (~zero_width) & (lower >= max)
        upper_bend |= (~zero_width) & (lower < max) & (upper > max) & (tautology if min is None else (lower > min))
    else:
        slope |= (~zero_width) & (lower >= min)

    if min is not None and max is not None:
        full_range |= (~zero_width) & (lower < min) & (upper > max)
        slope |= (~zero_width) & (lower >= min) & (upper <= max)

    return zero_width, flat_lower, flat_upper, slope, lower_bend, upper_bend, full_range


class BoundClamp(BoundActivation):
    def __init__(self, module, factory, adaptive_clamp=True, **kwargs):
        super().__init__(module, factory)
        self.adaptive_clamp = adaptive_clamp

        self.unstable_lower, self.unstable_slope_lower = None, None
        self.unstable_upper, self.unstable_slope_upper = None, None

    def clear_relaxation(self):
        super().clear_relaxation()
        self.unstable_lower, self.unstable_slope_lower = None, None
        self.unstable_upper, self.unstable_slope_upper = None, None

    @assert_bound_order
    def alpha_beta(self, preactivation):
        """
        Adaptive is similar to :BoundReLU: with the adaptivity being applied to both bends

        :param self:
        :param preactivation:
        """
        lower, upper = preactivation.lower, preactivation.upper
        zero_width, flat_lower, flat_upper, slope, lower_bend, upper_bend, full_range = regimes(lower, upper,
                                                                                                self.module.min,
                                                                                                self.module.max)

        self.alpha_lower, self.beta_lower = torch.zeros_like(lower), torch.zeros_like(lower)
        self.alpha_upper, self.beta_upper = torch.zeros_like(lower), torch.zeros_like(lower)

        act_lower, act_upper = self(lower), self(upper)

        # Use upper and lower in the bias to account for a small numerical difference between lower and upper
        # which ought to be negligible, but may still be present due to torch.isclose.
        if zero_width.any():
            self.alpha_lower[zero_width], self.beta_lower[zero_width] = 0, act_lower[zero_width]
            self.alpha_upper[zero_width], self.beta_upper[zero_width] = 0, act_upper[zero_width]

        min, max = self.module.min, self.module.max
        if min is not None and torch.is_tensor(min):
            min = min.view(*[1 for _ in range(self.alpha_lower.dim() - 1)], -1).expand_as(self.alpha_lower)

        if max is not None and torch.is_tensor(max):
            max = max.view(*[1 for _ in range(self.alpha_lower.dim() - 1)], -1).expand_as(self.alpha_lower)

        # Flat lower
        flat_lower_min = min[flat_lower] if min is not None and torch.is_tensor(min) else min
        self.alpha_lower[flat_lower], self.beta_lower[flat_lower] = 0, 0 if min is None else flat_lower_min
        self.alpha_upper[flat_lower], self.beta_upper[flat_lower] = 0, 0 if min is None else flat_lower_min

        # Flat upper
        flat_upper_max = max[flat_upper] if max is not None and torch.is_tensor(max) else max
        self.alpha_lower[flat_upper], self.beta_lower[flat_upper] = 0, 0 if max is None else flat_upper_max
        self.alpha_upper[flat_upper], self.beta_upper[flat_upper] = 0, 0 if max is None else flat_upper_max

        # Slope
        self.alpha_lower[slope], self.beta_lower[slope] = 1, 0
        self.alpha_upper[slope], self.beta_upper[slope] = 1, 0

        z = (self(upper) - self(lower)) / (upper - lower)

        # Lower bend
        if min is not None:
            if self.adaptive_clamp:
                # Utilize that bool->float conversion is true=1 and false=0
                a = (upper - min >= min - lower).to(lower.dtype)
            else:
                a = z

            lower_bend_min = min[lower_bend] if torch.is_tensor(min) else min
            self.alpha_lower[lower_bend], self.beta_lower[lower_bend] = a[lower_bend], lower_bend_min * (
                        1 - a[lower_bend])
            self.alpha_upper[lower_bend], self.beta_upper[lower_bend] = z[lower_bend], act_lower[lower_bend] - lower[
                lower_bend] * z[lower_bend]

            self.unstable_lower = lower_bend
            self.unstable_slope_lower = z[lower_bend].detach().clone().requires_grad_()

        # Upper bend
        if max is not None:
            if self.adaptive_clamp:
                # Utilize that bool->float conversion is true=1 and false=0
                a = (upper - max <= max - lower).to(lower.dtype)
            else:
                a = z

            upper_bend_max = max[upper_bend] if torch.is_tensor(max) else max
            self.alpha_lower[upper_bend], self.beta_lower[upper_bend] = z[upper_bend], act_lower[upper_bend] - lower[
                upper_bend] * z[upper_bend]
            self.alpha_upper[upper_bend], self.beta_upper[upper_bend] = a[upper_bend], upper_bend_max * (
                        1 - a[upper_bend])

            self.unstable_upper = upper_bend
            self.unstable_slope_upper = z[upper_bend].detach().clone().requires_grad_()

        # Full range
        if self.module.min is not None and self.module.max is not None:
            full_range_min = min[full_range] if torch.is_tensor(min) else min
            full_range_max = max[full_range] if torch.is_tensor(max) else max

            z_lower = (act_upper[full_range] - full_range_min) / (upper[full_range] - full_range_max)
            self.alpha_lower[full_range], self.beta_lower[full_range] = z_lower, act_upper[full_range] - upper[
                full_range] * z_lower

            z_upper = (full_range_max - act_lower[full_range]) / (full_range_max - lower[full_range])
            self.alpha_upper[full_range], self.beta_upper[full_range] = z_upper, act_lower[full_range] - lower[
                full_range] * z_upper

    def parameterize_alpha_beta(self, alpha_lower, alpha_upper, beta_lower, beta_upper):
        if self.unstable_lower is None and self.unstable_upper is None:
            logger.warning('Clamp bound not parameterized but expected to')

        if self.unstable_lower is not None:
            min = self.module.min
            if torch.is_tensor(min):
                min = min.view(*[1 for _ in range(alpha_lower.dim() - 1)], -1).expand_as(alpha_lower)[self.unstable_lower]
            alpha_lower[self.unstable_lower], beta_lower[self.unstable_lower] = self.unstable_slope_lower, min * (1 - self.unstable_slope_lower)

        if self.unstable_upper is not None:
            max = self.module.max
            if torch.is_tensor(max):
                max = max.view(*[1 for _ in range(alpha_upper.dim() - 1)], -1).expand_as(alpha_upper)[self.unstable_upper]
            alpha_upper[self.unstable_upper], beta_upper[self.unstable_upper] = self.unstable_slope_upper, max * (1 - self.unstable_slope_upper)

        return alpha_lower, alpha_upper, beta_lower, beta_upper

    def bound_parameters(self):
        if self.unstable_lower is None and self.unstable_upper is None:
            logger.warning('Clamp bound not parameterized but expected to')

        if self.unstable_lower is not None:
            yield self.unstable_slope_lower

        if self.unstable_upper is not None:
            yield self.unstable_slope_upper

    def clip_params(self):
        if self.unstable_lower is not None:
            clip_param_to_range_(self.unstable_slope_lower, (0, 1))

        if self.unstable_upper is not None:
            clip_param_to_range_(self.unstable_slope_upper, (0, 1))

    def project_grads(self):
        if self.unstable_lower is not None:
            proj_grad_to_range_(self.unstable_slope_lower, (0, 1))

        if self.unstable_upper is not None:
            proj_grad_to_range_(self.unstable_slope_upper, (0, 1))



import abc
import logging

import torch
from torch import nn
from torch.optim import Adam

from .bounds import LinearBounds, IntervalBounds


logger = logging.getLogger(__name__)


class BoundModule(nn.Module, abc.ABC):
    def __init__(self, module, factory, **kwargs):
        super().__init__()
        self.module = module

        self.alpha_optimizer = kwargs.get('alpha_optimizer', Adam)
        self.alpha_optimizer_params = kwargs.get('alpha_optimizer_params', {'lr': 1e-2, 'betas': (0.5, 0.99)})
        self.alpha_iterations = kwargs.get('alpha_iterations', 20)

    @torch.no_grad()
    def crown_relax(self, region):
        # Force bounds based on IBP, which may be tighter. More importantly, this also works if say there
        # is a Clamp in front of a Log which requires x > 0, and a linear relaxation may violate this
        interval_bounds = IntervalBounds(region, region.lower, region.upper)
        self.ibp_forward(interval_bounds, save_input_bounds=True)

        # No grad for relaxations improves accuracy and stabilizes training for CROWN.
        while self.need_relaxation:
            linear_bounds, module, *extra = self.backward_relaxation(region)
            module.set_relaxation(linear_bounds, *extra)

    @torch.no_grad()
    def ibp_relax(self, region):
        # No grad for relaxations improves accuracy. CROWN-IBP is already stable in training.
        bounds = IntervalBounds(region, region.lower, region.upper)
        self.ibp_forward(bounds, save_relaxation=True)

    def crown(self, region, bound_lower=True, bound_upper=True, alpha=False):
        return self.crown_with_relaxation(self.crown_relax, region, bound_lower, bound_upper, alpha)

    def crown_ibp(self, region, bound_lower=True, bound_upper=True, alpha=False):
        return self.crown_with_relaxation(self.ibp_relax, region, bound_lower, bound_upper, alpha)

    def crown_with_relaxation(self, relax, region, bound_lower, bound_upper, alpha):
        out_size = self.propagate_size(region.lower.size(-1))

        relax(region)

        if alpha:
            linear_bounds = self.alpha_crown(region, out_size, bound_lower, bound_upper)
        else:
            linear_bounds = self.initial_linear_bounds(region, out_size, lower=bound_lower, upper=bound_upper)
            linear_bounds = self.crown_backward(linear_bounds, False)

        self.clear_relaxation()
        return linear_bounds

    def initial_linear_bounds(self, region, out_size, lower=True, upper=True):
        device, dtype = region.lower.device, region.lower.dtype
        W_tilde = torch.eye(out_size, device=device, dtype=dtype).unsqueeze(-3).expand(*region.lower.size()[:-1], out_size, out_size)
        bias = torch.zeros((out_size,), device=device, dtype=dtype).unsqueeze(-2).expand(*region.lower.size()[:-1], out_size)

        lower = (W_tilde, bias) if lower else None
        upper = (W_tilde, bias) if upper else None

        linear_bounds = LinearBounds(region, lower, upper)
        return linear_bounds

    def alpha_crown(self, region, out_size, bound_lower, bound_upper):
        params = list(self.bound_parameters())

        if all([param.numel() == 0 for param in params]):
            logger.warning('No parameters available for alpha-CROWN. Check architecture of network.')

            linear_bounds = self.initial_linear_bounds(region, out_size, lower=bound_lower, upper=bound_upper)
            return self.crown_backward(linear_bounds, False)

        self.optimize_bounds(region, out_size, params, bound_lower=bound_lower, bound_upper=bound_upper)

        linear_bounds = self.initial_linear_bounds(region, out_size, lower=bound_lower, upper=bound_upper)
        return self.crown_backward(linear_bounds, True)

    @torch.enable_grad()
    def optimize_bounds(self, region, out_size, params, bound_lower=True, bound_upper=True):
        optimizer = self.alpha_optimizer(params, **self.alpha_optimizer_params)

        for iteration in range(self.alpha_iterations):
            linear_bounds = self.initial_linear_bounds(region, out_size, lower=bound_lower, upper=bound_upper)
            linear_bounds = self.crown_backward(linear_bounds, True)

            optimizer.zero_grad(set_to_none=True)  # Bound parameters
            self.zero_grad(set_to_none=True)       # Network/model parameters

            loss = self.alpha_loss(linear_bounds, bound_lower, bound_upper)
            loss.backward()
            self.project_grads()
            optimizer.step()

            self.clip_params()  # Projected Gradient Descent

        optimizer.zero_grad(set_to_none=True)
        self.zero_grad(set_to_none=True)

    def alpha_loss(self, linear_bounds, bound_lower, bound_upper):
        # Let this be a separate function to allow subclasses to specialize loss calculation.
        # Useful if other goals (e.g. distance between bounds) are to be optimized instead.

        interval_bounds = linear_bounds.concretize()

        loss = 0
        if bound_lower:
            # Use sum for aggregation because bound parameters are per sample, hence we need to scale the loss with
            # the number of samples
            loss = loss - interval_bounds.lower.sum()

        if bound_upper:
            loss = loss + interval_bounds.upper.sum()

        return loss

    @property
    @abc.abstractmethod
    def need_relaxation(self):
        raise NotImplementedError()

    def clear_relaxation(self):
        pass

    def set_relaxation(self, linear_bounds, *args):
        pass

    def backward_relaxation(self, region):
        pass

    @abc.abstractmethod
    def crown_backward(self, linear_bounds, optimize):
        raise NotImplementedError()

    def ibp(self, region):
        bounds = IntervalBounds(region, region.lower, region.upper)
        return self.ibp_forward(bounds)

    @abc.abstractmethod
    def ibp_forward(self, bounds, save_relaxation=False, save_input_bounds=False):
        raise NotImplementedError()

    @abc.abstractmethod
    def propagate_size(self, in_size):
        raise NotImplementedError()

    def forward(self, *args, **kwargs):
        return self.module(*args, **kwargs)

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        return self.module.state_dict(destination=destination, prefix=prefix, keep_vars=keep_vars)

    def load_state_dict(self, state_dict: 'OrderedDict[str, Tensor]', strict: bool = True):
        # Assume that submodules are constructed with the factory, since that produces
        # references down the bound module graph corresponding to the model graph.
        # Then loading the state dict at the top level automatically populates the weights in all sub-bound modules.
        self.module.load_state_dict(state_dict=state_dict, strict=strict)

    def bound_parameters(self):
        return []

    def clip_params(self):
        pass

    def project_grads(self):
        pass
